#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import app
from flask import render_template, url_for, redirect, flash, request
import fastr
import fastr.exceptions
import requests

def shutdown_server():
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()


@app.route('/shutdown')
def shutdown():
    shutdown_server()
    return """
    <!DOCTYPE html>
    <html lang="en">
      <head>
        <meta charset="utf-8">
        <title>title</title>
        <link rel="stylesheet" href="style.css">
      </head>
      <body>
        Server shutting down...
      </body>
    </html>
    """

@app.route('/')
@app.route('/index')
def index():
    return render_template('index.html')


@app.route('/websocketclient')
def websocket_client():
    return render_template('websocket_client.html')


@app.route('/networks')
def networks():
    # http://stackoverflow.com/questions/26868372/calling-rest-api-from-the-same-server
    # http://stackoverflow.com/questions/20169326/calling-flask-restful-api-resource-methods
    # import api

    # print requests.get("http://localhost:9000/api/networks/")
    # test = api.api.resource("/networks/")
    # print(request.args.get("/api/networks/"))
    networks = fastr.networklist.values()

    return render_template('networks.html')


@app.route('/network/<name>')
def network(name=None):
    return render_template('network.html', name=name)


@app.route('/doc', endpoint='docs')
def doc():
    return render_template('docs.html', url=url_for('static', filename='doc/index.html'))


@app.route('/prov')
def prov():
    return render_template('prov.html')


@app.route('/tool')
@app.route('/tool/<toolname>')
@app.route('/tool/<toolname>/<version>')
def tool(toolname=None, version=None):
    if toolname is None and version is None:
        unique_toolnames = set(['{}.{}'.format(key[0], key[1]) for key in fastr.toollist.keys()])
        print('Unique tool names: {}'.format(unique_toolnames))
        toollist = sorted([(tool, sorted([str(v) for v in fastr.toollist.toolversions(tool)], reverse=True)) for tool in unique_toolnames])
        return render_template('tool_overview.html', tools=toollist)

    # If either a toolname or both version and toolname is set:
    if toolname in fastr.toollist:
        versions = sorted([str(v) for v in fastr.toollist.toolversions(toolname)], reverse=True)
        tools = [(v, fastr.toollist[toolname, v]) for v in versions]
    else:
        flash("Tool ({}) is not known.".format(toolname), 'danger')
        tools = []
    # When there is no explicit version set, appoint the first version in the list as selected.
    if version is None and tools != []:
        version = tools[0][0]
    return render_template('tool.html', toolname=toolname, selected_version=version, tools=tools)



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Divide a number by another number')
    parser.add_argument('--in1', metavar='IN1', dest='in1', type=float, nargs='+', required=True, help='First number (numerator)')
    parser.add_argument('--in2', metavar='IN2', dest='in2', type=float, nargs='+', required=True, help='Second number (denominator)')

    args = parser.parse_args()

    if len(args.in1) != len(args.in2):
        raise ValueError("Number of value in each argument should match!")

    result = [x / y for x, y in zip(args.in1, args.in2)]

    print('RESULT={}'.format(json.dumps(result)))



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the manager class for IOPlugins and the
base class for all IOPlugins
"""

import argparse
import json
import os
import urlparse as up

if __name__ == '__main__':
    FASTR_LOG_TYPE = 'console'

import fastr
from fastr.core.baseplugin import Plugin
from fastr.core.datatypemanager import typelist
from fastr.core.pluginmanager import PluginSubManager
from fastr.core.tool import Tool
from fastr.core.version import Version
import fastr.data.url as urltools
from fastr.datatypes import URLType, TypeGroup
import fastr.exceptions as exceptions
from abc import abstractproperty
from abc import ABCMeta


class IOPluginManager(PluginSubManager):
    """
    A mapping containing the IOPlugins known to this system
    """

    def __init__(self):
        """
        Create the IOPluginManager and populate it.

        :return: newly created IOPluginManager
        """
        super(IOPluginManager, self).__init__(parent=fastr.plugin_manager,
                                              plugin_class=self.plugin_class)
        self._key_map = {}

    def cleanup(self):
        """
        Cleanup all plugins, this closes files, connections and other things
        that could be left dangling otherwise.
        """
        for ioplugin in self.values():
            ioplugin.cleanup()

    @property
    def plugin_class(self):
        """
        The PluginClass of the items of the BasePluginManager
        """
        return IOPlugin

    def __keytransform__(self, key):
        try:
            return self._key_map[key]
        except KeyError:
            self._key_map.clear()
            for id_, value in self.data.items():
                if isinstance(value.scheme, tuple):
                    for scheme in value.scheme:
                        self._key_map[scheme] = id_
                else:
                    self._key_map[value.scheme] = id_

            return self._key_map[key]

    def __iter__(self):
        for value in self.data.values():
            if isinstance(value.scheme, tuple):
                for scheme in value.scheme:
                    yield scheme
            else:
                yield value.scheme

    def _print_key(self, key):
        """
        Get a printable string for the IOPlugin key

        :param key: key to get the printable version for
        :return: printable version of the key
        :rtype: str
        """
        return self[key].status.value, '{}://'.format(key)

    def expand_url(self, url):
        """
        Expand the url by filling the wildcards. This function checks the url scheme
        and uses the expand function of the correct IOPlugin.

        :param str url: url to expand
        :return: list of urls
        :rtype: list of str
        """
        parsed_url = up.urlparse(url)
        return self[parsed_url.scheme].expand_url(url)

    def pull_source_data(self, url, outdir, sample_id, datatype=None):
        """
        Retrieve data from an external source. This function checks the url scheme and
        selects the correct IOPlugin to retrieve the data.

        :param url: url to pull
        :param str outdir: the directory to write the data to
        :param DataType datatype: the datatype of the data, used for determining
                                  the total contents of the transfer
        :return: None
        """
        parsed_url = up.urlparse(url)
        return self[parsed_url.scheme].pull_source_data(url, outdir, sample_id, datatype)

    def put_url(self, inpath, outurl):
        """
        Put the files to the external data store.

        :param inpath: path to the local data
        :param outurl: url to where to store the data in the external data store.
        """
        parsed_url = up.urlparse(outurl)
        return self[parsed_url.scheme].put_url(inpath, outurl)

    def url_to_path(self, url):
        """
        Retrieve the path for a given url

        :param str url: the url to parse
        :return: the path corresponding to the input url
        :rtype: str
        """
        parsed_url = up.urlparse(url)
        return self[parsed_url.scheme].url_to_path(url)

    @staticmethod
    def register_url_scheme(scheme):
        """
        Register a custom scheme to behave http like. This is needed to parse
        all things properly with urlparse.

        :param scheme: the scheme to register
        """
        for method in [s for s in dir(up) if s.startswith('uses_')]:
            if scheme not in getattr(up, method):
                getattr(up, method).append(scheme)

    def populate(self):
        """
        Populate the IOPlugins manager. After the default directory scan, add
        the vfs IOPlugin and create the Tools for the IOPlugins
        """
        super(IOPluginManager, self).populate()

    @staticmethod
    def create_ioplugin_tool():
        """
        Create the tools which handles sinks and sources. The command of this tool is the main of core.ioplugin.
        """
        # First create the Enum for the behaviour input
        fastr.typelist.create_enumtype('__ioplugin__behaviour__Enum__', ('source', 'sink'))

        if ('fastr', 'Source', '1.0') not in fastr.toollist:
            source_tool = Tool()
            source_tool.id = 'Source'
            source_tool.name = source_tool.id
            source_tool.version = Version('1.0')
            source_tool.namespace = 'fastr'
            source_tool.node_class = 'Node'
            source_tool.filename = __file__
            source_tool.tests = []
            source_tool._target = None
            source_tool.command = {'authors': ['fastr devs'],
                                   'description': 'Source Tool: tool to handle SourceNode operation.',
                                   'targets': [{'os': '*', 'arch': '*', 'bin': 'ioplugin.py', 'interpreter': 'python', 'paths': './', 'module': None}],
                                   'url': 'http://www.bigr.nl/fastr',
                                   'version': Version('1.0')}

            input_params = [{'id': 'input',
                             'name': 'input',
                             'description': 'The data to be retrieved by the SourceNode',
                             'datatype': 'String',
                             'prefix': '--input',
                             'repeat_prefix': False,
                             'cardinality': '1-*',
                             'required': True,
                             'hidden': True},
                            {'id': 'behaviour',
                             'name': 'behaviour',
                             'enum': ('source', 'sink'),
                             'prefix': '--behaviour',
                             'repeat_prefix': False,
                             'cardinality': '1',
                             'required': True,
                             'hidden': True},
                            {'id': 'datatype',
                             'name': 'datatype',
                             'datatype': 'String',
                             'prefix': '--datatype',
                             'cardinality': '1',
                             'required': True,
                             'hidden': True},
                            {'id': 'targetdir',
                             'name': 'targetdir',
                             'description': 'The location to store the result',
                             'datatype': 'Directory',
                             'prefix': '--output',
                             'cardinality': '1',
                             'required': True,
                             'hidden': True},
                            {'id': 'sample_id',
                             'name': 'sample_id',
                             'description': 'The sample id for the SourceJob that is run',
                             'datatype': 'String',
                             'prefix': '--sample_id',
                             'cardinality': '1',
                             'required': True,
                             'hidden': True}]

            output_params = [{'id': 'output',
                              'name': 'The output urls in vfs scheme',
                              'description': '',
                              'datatype': 'AnyType',
                              'cardinality': 'unknown',
                              'automatic': True,
                              'location': '^__IOPLUGIN_OUT__=(.*)$',
                              'method': 'json'}]

            interface = {'inputs': input_params, 'outputs': output_params}
            source_tool.interface = fastr.interfaces['FastrInterface'](id_='source-interface', document=interface)

            # Register source tool with the ToolManager
            fastr.toollist[source_tool.namespace, source_tool.id, source_tool.version] = source_tool

        if ('fastr', 'Sink', '1.0') not in fastr.toollist:
            sink_tool = Tool()
            sink_tool.id = 'Sink'
            sink_tool.name = sink_tool.id
            sink_tool.version = Version('1.0')
            sink_tool.namespace = 'fastr'
            sink_tool.node_class = 'Node'
            sink_tool.filename = __file__
            sink_tool.tests = []
            sink_tool._target = None
            sink_tool.command = {'authors': ['fastr devs'],
                                 'description': 'Sink Tool: tool to handle Sink Node operation.',
                                 'targets': [{'os': '*', 'arch': '*', 'bin': 'ioplugin.py', 'interpreter': 'python', 'paths': './', 'module': None}],
                                 'url': 'http://www.bigr.nl/fastr',
                                 'version': Version('1.0')}

            input_params = [{'id': 'input',
                             'name': 'The url to process (can also be a list)',
                             'description': 'The data to be store by the SinkNode',
                             'datatype': 'AnyType',
                             'prefix': '--input',
                             'cardinality': '1-*',
                             'required': True, },
                            {'id': 'behaviour',
                             'name': 'behaviour',
                             'enum': ('source', 'sink'),
                             'prefix': '--behaviour',
                             'cardinality': '1',
                             'required': True,
                             'hidden': True},
                            {'id': 'output',
                             'name': 'output',
                             'datatype': 'String',
                             'prefix': '--output',
                             'cardinality': '1',
                             'required': True,
                             'hidden': True},
                            {'id': 'datatype',
                             'name': 'datatype',
                             'datatype': 'String',
                             'prefix': '--datatype',
                             'cardinality': '1',
                             'required': True,
                             'hidden': True},
                            ]

            output_params = []

            interface = {'inputs': input_params, 'outputs': output_params}
            sink_tool.interface = fastr.interfaces['FastrInterface'](id_='source-interface', document=interface)

            # Register sink tool with the ToolManager
            fastr.toollist[sink_tool.namespace, sink_tool.id, sink_tool.version] = sink_tool


class IOPlugin(Plugin):
    """
    :py:class:`IOPlugins <fastr.core.ioplugin.IOPlugin>` are used for data import
    and export for the sources and sinks. The main use of the
    :py:class:`IOPlugins <fastr.core.ioplugin.IOPlugin>` is during execution (see
    :ref:`Execution <manual_execution>`). The :py:class:`IOPlugins <fastr.core.ioplugin.IOPlugin>`
    can be accessed via ``fastr.ioplugins``, but generally there should be no need
    for direct interaction with these objects. The use of is mainly via the URL
    used to specify source and sink data.
    """
    __metaclass__ = ABCMeta
    _instantiate = True

    def __init__(self):
        """
        Initialization for the IOPlugin

        :return: newly created IOPlugin
        """
        super(IOPlugin, self).__init__()
        self._results = {}

    @abstractproperty
    def scheme(self):
        """
        ``(abstract)`` This abstract property is to be overwritten by a subclass to indicate
        the url scheme associated with the IOPlugin.
        """
        raise exceptions.FastrNotImplementedError("IOPlugin scheme is not set")

    def url_to_path(self, url):
        """
        ``(abstract)`` Get the path to a file from a url.

        :param str url: the url to retrieve the path for
        :return: the corresponding path
        :rtype: str
        """
        # This is a placeholder function, so we do no use our arguments
        # pylint: disable=unused-argument
        raise exceptions.FastrNotImplementedError('{} is not for working with urls'.format(self.scheme))

    def fetch_url(self, inurl, outfile):
        """
        ``(abstract)``  Fetch a file from an external data source.

        :param inurl: url to the item in the data store
        :param outpath: path where to store the fetch data locally
        """
        # This is a placeholder function, so we do no use our arguments
        # pylint: disable=unused-argument
        raise exceptions.FastrNotImplementedError('{} is not for direct url data retrieval'.format(self.scheme))

    def fetch_value(self, inurl):
        """
        ``(abstract)``  Fetch a value from an external data source.

        :param inurl: the url of the value to retrieve
        :return: the fetched value
        """
        # This is a placeholder function, so we do no use our arguments
        # pylint: disable=unused-argument
        raise exceptions.FastrNotImplementedError('{} is not for direct value data retrieval'.format(self.scheme))

    def put_url(self, inpath, outurl):
        """
        ``(abstract)`` Put the files to the external data store.

        :param inpath: path to the local data
        :param outurl: url to where to store the data in the external data store.
        """
        # This is a placeholder function, so we do no use our arguments
        # pylint: disable=unused-argument
        raise exceptions.FastrNotImplementedError('{} is not for direct url data storage'.format(self.scheme))

    def put_value(self, value, outurl):
        """
        ``(abstract)`` Put the files to the external data store.

        :param value: the value to store
        :param outurl: url to where to store the data in the external data store.
        """
        # This is a placeholder function, so we do no use our arguments
        # pylint: disable=unused-argument
        raise exceptions.FastrNotImplementedError('{} is not for direct value data storage'.format(self.scheme))

    def path_to_url(self, path, mountpoint=None):
        """
        ``(abstract)`` Construct an url from a given mount point and a relative
        path to the mount point.

        :param str path: the path to determine the url for
        :param mountpoint: the mount point to use, will be automatically
                           detected if None is given
        :type mountpoint: str or None
        :return: url matching the path
        :rtype: str
        """
        # This is a placeholder function, so we do no use our arguments
        # pylint: disable=unused-argument
        raise exceptions.FastrNotImplementedError('{} is not for working with urls'.format(self.scheme))

    def setup(self, *args, **kwargs):
        """
        ``(abstract)`` Setup before data transfer. This can be any function
        that needs to be used to prepare the plugin for data transfer.
        """
        # This is a placeholder function, so we do no use our arguments
        # pylint: disable=unused-argument,no-self-use
        pass

    def cleanup(self):
        """
        ``(abstract)`` Clean up the IOPlugin. This is to do things like
        closing files or connections. Will be called when the plugin is no
        longer required.
        """
        # This is a placeholder function, so we do no use our arguments
        # pylint: disable=unused-argument,no-self-use
        pass

    def expand_url(self, url):
        """
        ``(abstract)`` Expand an URL. This allows a source to collect multiple
        samples from a single url. The URL will have a wildcard or point to
        something with info and multiple urls will be returned.

        :param str url: url to expand
        :return: the resulting url(s), a tuple if multiple, otherwise a str
        :rtype: str or tuple of str
        """
        # This is a placeholder function, so we do no use our arguments
        # pylint: disable=unused-argument,no-self-use
        return url

    @staticmethod
    def isurl(string):
        """
        Test if given string is an url.

        :param str string: string to test
        :return: ``True`` if the string is an url, ``False`` otherwise
        :rtype: bool
        """
        parsed_url = up.urlparse(str(string))
        return parsed_url.scheme != ''

    @staticmethod
    def print_result(result):
        """
        Print the result of the IOPlugin to stdout to be picked up by the tool

        :param result: value to print as a result
        :return: None
        """
        print('__IOPLUGIN_OUT__={}'.format(json.dumps(result)))

    def pull_source_data(self, inurl, outdir, sample_id, datatype=None):
        """
        Transfer the source data from inurl to be available in outdir.

        :param str inurl: the input url to fetch data from
        :param str outdir: the directory to write the data to
        :param DataType datatype: the datatype of the data, used for determining
                                  the total contents of the transfer
        :return: None
        """
        results = {}
        self.setup()

        # First expand the URL
        valuelist = self.expand_url(inurl)

        fastr.log.debug('[{}] pulling sample {} with value {} and datatype {}'.format(self.scheme, sample_id, inurl, datatype))

        if isinstance(valuelist, tuple):
            # We expanded the URL, so now process each new value/URL seperately
            if len(valuelist) == 0:
                raise exceptions.FastrValueError(('No data found when expanding'
                                                  ' URL {}, this probably means '
                                                  'the URL is not correct.').format(inurl))
            for cardinality_nr, (sub_sample_id, value) in enumerate(valuelist):
                if sub_sample_id is None:
                    fastr.log.debug('Changing sub sample id from None to {}'.format(sub_sample_id))
                    sub_sample_id = '{}_{}'.format(sample_id, cardinality_nr)
                fastr.log.debug('Found expanded item {}: {}'.format(sub_sample_id, value))
                if self.isurl(value):
                    # Expanded value is an URL, so it need to be processed
                    outsubdir = os.path.join(outdir, str(sub_sample_id))
                    if not os.path.isdir(outsubdir):
                        os.mkdir(outsubdir)
                    result = fastr.ioplugins.pull_source_data(value, outsubdir, sub_sample_id, datatype)
                    results.update(result)
                else:
                    # Expanded value is a value, so we assume this is the value to be used
                    results[sub_sample_id] = (value,)
        elif isinstance(valuelist, str):
            # The expand did not change the URL
            if valuelist != inurl:
                raise exceptions.FastrValueError('If valuelist is a str, it should be the original inurl!')

            # Check against None to avoid issubclass throwing an error
            if datatype is not None and issubclass(datatype, TypeGroup):
                if all(issubclass(x, URLType) for x in datatype.members):
                    urltype = True
                elif all(not issubclass(x, URLType) for x in datatype.members):
                    urltype = False
                else:
                    raise exceptions.FastrNotImplementedError('Cannot use a SourceNode that can supply mixed URL and Value types')
            else:
                # Check against None to avoid issubclass throwing an error
                urltype = datatype is not None and issubclass(datatype, URLType)

            fastr.log.debug('[{}] the urltype found is {}'.format(self.scheme, urltype))

            if urltype:
                outfile = os.path.join(outdir, urltools.basename(inurl))
                result = self.fetch_url(inurl, outfile)

                if not result:
                    raise exceptions.FastrIOError('Could not retrieve data from {}'.format(inurl))

                if datatype is None or issubclass(datatype, TypeGroup):
                    datatype = typelist.guess_type(result, options=datatype)
                    fastr.log.debug('Refined datatype to {}'.format(datatype))

                if datatype is not None:
                    contents = datatype.content(inurl, result)
                    fastr.log.debug('Found contents {}'.format(contents))
                else:
                    contents = [(inurl, result)]

                for extrain, extraout in contents:
                    if extrain == inurl and extraout == result:
                        fastr.log.info('Skipping original file {} -> {}'.format(extrain, extraout))
                        continue

                    fastr.log.debug('Orig {} -> {}'.format(inurl, result))
                    fastr.log.debug('Processing original file {} -> {}'.format(extrain, extraout))

                    if not os.path.exists(extraout):
                        extra_result = self.fetch_url(extrain, extraout)
                        if not extra_result:
                            raise exceptions.FastrIOError('Could not retrieve data from {} to {}'.format(extrain, extraout))

                results[sample_id] = (result,)
            else:
                result = self.fetch_value(inurl)
                results[sample_id] = (result,)
        else:
            fastr.log.error('Expand of {}  returned an invalid type! ({} after expansion)'.format(inurl, valuelist))

        return results

    def push_sink_data(self, inpath, outurl, datatype=None):
        """
        Write out the sink data from the inpath to the outurl.

        :param str inpath: the path of the data to be pushed
        :param str outurl: the url to write the data to
        :param DataType datatype: the datatype of the data, used for determining
                                  the total contents of the transfer
        :return: None
        """
        fastr.log.info('Push sink called with: {}, {}'.format(inpath, outurl))
        self.setup()

        if datatype is None or issubclass(datatype, TypeGroup):
            previous_datatype = datatype.id if datatype is not None else None
            datatype = typelist.guess_type(inpath, options=datatype)
            fastr.log.info('Determined specific datatype as {} (based on {})'.format(datatype.id, previous_datatype))

        if datatype is not None and issubclass(datatype, URLType):
            contents = datatype.content(inpath, outurl)
            fastr.log.info('Found URL contents: {}'.format(contents))
        else:
            contents = [(inpath, outurl)]
            fastr.log.info('Found value/simple contents: {}'.format(contents))

        for extracontent, extratarget in contents:
            if datatype is not None and issubclass(datatype, URLType):
                result = self.put_url(extracontent, extratarget)
            else:
                result = self.put_value(extracontent, extratarget)
            if not result:
                raise exceptions.FastrIOError('Could not store data from {} to {}'.format(extracontent, extratarget))

    @staticmethod
    def _correct_separators(path):
        """
        Translates the URL seperator '/' into the apropriate seperator for the OS

        :param str path: the path to correct
        :return: path with corrected separators
        :rtype: str
        """
        return path.replace('/', os.path.sep)


def main():
    """
    The main entry point for command line access to the IOPlugin
    """
    # The main function creates an entry point for the SourceNode and SinkNode
    parser = argparse.ArgumentParser(description="executes an ioplugin",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-b', '--behaviour', type=str, required=True,
                        choices=['source', 'sink'],
                        help="Select if the script should behave like"
                             " a 'source' or a 'sink'")
    parser.add_argument('-i', '--input', nargs='*', type=str, required=True,
                        help="The url to process (can also be a list)")
    parser.add_argument('-o', '--output', type=str, metavar='OUTPUT', required=True,
                        help="The output urls in vfs scheme (can also be a"
                             " list and should be the same size as --inurl)")
    parser.add_argument('-d', '--datatype', default=None, type=str,
                        help="The datatype of the source/sink data to handle")
    parser.add_argument('-s', '--sample_id', default=None, type=str,
                        help="The sample_id of the source/sink data to handle")

    args = parser.parse_args()

    fastr.log.info(('Arguments:\nbehaviour: {}\ninput: {}\noutput: {}\n'
                    'datatype: {}\nsample_id: {}').format(args.behaviour,
                                                          args.input,
                                                          args.output,
                                                          args.datatype,
                                                          args.sample_id))

    inputs = args.input
    output = args.output
    sample_id = args.sample_id

    # Get the datatype to use
    datatype = args.datatype
    if datatype is not None:
        datatype = fastr.typelist[datatype]
    fastr.log.debug("datatype to use: {}".format(datatype))

    if args.behaviour == 'source':
        if sample_id is None:
            message = 'For Source behaviour the sample id needs to be defined!'
            fastr.log.critical(message)
            raise ValueError(message)

        if urltools.isurl(output):
            message = 'For Source behaviour the output should be a directory, not a URL! (found {})'.format(output)
            fastr.log.critical(message)
            raise ValueError(message)

        results = []
        for input_ in inputs:
            if urltools.isurl(input_):
                parsed_input = up.urlparse(input_)
                try:
                    plugin = fastr.ioplugins[parsed_input.scheme]
                except KeyError:
                    message = "No valid scheme is supplied in: {} (Found {}). The following schemes are supported: {}".format(input_, parsed_input.scheme, ' '.join(fastr.ioplugins.keys()))
                    fastr.log.error(message)
                    raise exceptions.FastrUnknownURLSchemeError(message)

                results.append(plugin.pull_source_data(input_, output, sample_id, datatype))
            else:
                results.append(input_)

        if len(results) == 1:
            # If we have only one result, there is no need to merge
            result = results[0]
        else:
            # Merge multiple cardinality into a single result
            samples = set()
            for part in results:
                # Check if we got value or if the value was a simple value
                if isinstance(part, dict):
                    samples.update(part.keys())

            # For each sample collect all found values
            result = {}
            for sample_id in samples:
                result[sample_id] = temp = []

                for part in results:
                    if isinstance(part, dict):
                        # We have a dictionary with {sampleid: (value1, value2)} format
                        # FIXME: Is this behaviour desired? It might lead to strange results
                        temp.extend(part.get(sample_id, []))
                    else:
                        # We have a value (not URL) that was given, we assume this to be used for ALL samples (broadcasting)
                        temp.append(part)

        IOPlugin.print_result(result)
    elif args.behaviour == 'sink':
        if len(inputs) != 1:
            message = 'Sink can only handle one input per job (found: {})'.format(inputs)
            fastr.log.critical(message)
            raise ValueError(message)

        input_ = inputs[0]

        if urltools.isurl(input_):
            message = 'For Sink behaviour the sink should be a path, not a URL! (found {})'.format(input_)
            fastr.log.critical(message)
            raise ValueError(message)

        parsed_output = up.urlparse(output)
        try:
            plugin = fastr.ioplugins[parsed_output.scheme]
        except KeyError:
            message = "No valid scheme is supplied in: {} (Found {}). The following schemes are supported: {}".format(output, parsed_output.scheme, ' '.join(fastr.ioplugins.keys()))
            fastr.log.error(message)
            raise exceptions.FastrUnknownURLSchemeError(message)

        plugin.push_sink_data(input_, output, datatype)

    # Cleanup all plugins
    fastr.ioplugins.cleanup()


if __name__ == '__main__':
    main()



import os
import fastr

from fastr.core.node import MacroNode


def create_macro_network():
    # Create Network
    network = fastr.Network('add_multiple_ints')
    # Create first source
    input_ = network.create_source(fastr.typelist['Int'], id_='input')

    # Create calculation nodes
    addint1 = network.create_node(fastr.toollist['AddInt'], id_='addint1')
    addint2 = network.create_node(fastr.toollist['AddInt'], id_='addint2')
    # Link network
    addint1.inputs['left_hand'] = input_.output
    addint1.inputs['right_hand'] = 10,
    addint2.inputs['left_hand'] = addint1.outputs['result']
    addint2.inputs['right_hand'] = 100,
    # Create a sink to save the data
    sink = network.create_sink(fastr.typelist['Int'], id_='macro_sink')
    # Link the addint node to the sink
    sink.inputs['input'] = addint2.outputs['result']
    #network.draw_network(name=network.id, img_format='png')
    return network


def create_super_macro_node():
    network = fastr.Network('macro_container')
    # Create Outputs
    input_value = network.create_source(fastr.typelist['Int'], id_='input_value')
    # Create Macro Networks
    macro_network_1 = create_macro_network()
    macro_network_2 = create_macro_network()
    add_multiple_ints_node_1 = network.create_macro(macro_network_1, id_='add_multiple_ints_1')
    add_multiple_ints_node_2 = network.create_macro(macro_network_2, id_='add_multiple_ints_2')
    # Create Sink
    output_value = network.create_sink(fastr.typelist['Int'], id_='output_value')
    # Create Links
    add_multiple_ints_node_1.inputs['input'] = input_value.output
    add_multiple_ints_node_2.inputs['input'] = add_multiple_ints_node_1.outputs['macro_sink']
    output_value.inputs['input'] = add_multiple_ints_node_2.outputs['macro_sink']
    #network.draw_network(name=network.id, img_format='png')
    return network


def create_network():
    macro_network = create_super_macro_node()

    # Create Network
    test_network = fastr.Network('test_macro_node')
    # Create data source
    input_ = test_network.create_source(fastr.typelist['Int'], id_='source')
    # Create MacroNode
    add_multiple_ints_node = test_network.create_macro(macro_network, id_='add_multiple_ints')

    # Create sink
    sink = test_network.create_sink(fastr.typelist['Int'], id_='test_network_sink')

    # Adjust constants(non required inputs) in macro network
    #  add_multiple_ints_node.inputs['const_addint1_value2__add_multiple_ints_1'] = 1234, #input.output
    # Link the network
    add_multiple_ints_node.inputs['input_value'] = input_.output
    sink.inputs['input'] = add_multiple_ints_node.outputs['output_value']

    return test_network


def source_data(network):
    return {'source': [1, 'vfslist://example_data/add_ints/values']}


def sink_data(network):
    return {'test_network_sink': 'vfs://tmp/results/{}/result_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id)}


def main():
    network = create_network()

    # Validate and execute network
    if network.is_valid():
        network.draw_network(name=network.id, img_format='png')
        network.execute(source_data(network), sink_data(network))
    else:
        print "Network was not valid"


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
import sys
import threading
import time
import traceback

import fastr
import fastr.resources
from fastr.core.baseplugin import PluginState
from fastr.execution.executionpluginmanager import ExecutionPlugin, JobAction
from fastr.execution.job import Job, JobState
from fastr.utils.classproperty import classproperty

try:
    from rq import Queue
    from redis import Redis
    IMPORT_SUCCESS = True
except ImportError:
    IMPORT_SUCCESS = False


class RQExecution(ExecutionPlugin):
    """
    A execution plugin based on Redis Queue. Fastr will submit jobs to the
    redis queue and workers will peel the jobs from the queue and process
    them.

    This system requires a running redis database and the database url has to
    be set in the fastr configuration.

    .. note::

        This execution plugin required the ``redis`` and ``rq`` packages to
        be installed before it can be loaded properly.
    """
    if not IMPORT_SUCCESS:
        _status = (PluginState.failed, 'Could not load rq and/or redis!')

    def __init__(self, finished_callback=None, cancelled_callback=None, status_callback=None):
        super(RQExecution, self).__init__(finished_callback, cancelled_callback, status_callback)
        redis = Redis.from_url(fastr.config.rq_host)
        self.queue = Queue(name=fastr.config.rq_queue, connection=redis, default_timeout=-1)
        self.rq_jobs = {}
        self.running = True

        fastr.log.debug('Creating rq job collector')
        self.collector = threading.Thread(name='RQJobCollector-0', target=self.check_finished, args=())
        self.collector.daemon = True
        fastr.log.debug('Starting rq job collector')
        self.collector.start()

    @classproperty
    def configuration_fields(cls):
        return {
            "rq_host": (str, "redis://localhost:6379/0", "The url of the redis serving the redis queue"),
            "rq_queue": (str, "default", "The redis queue to use"),
        }

    def cleanup(self):
        super(RQExecution, self).cleanup()

    def _job_finished(self, result):
        pass

    def _cancel_job(self, job):
        pass

    def _release_job(self, job):
        if not isinstance(job, Job):
            job = self.job_dict[job]
        self.queue_job(job)

    def _queue_job(self, job):
        action = self.check_job_requirements(job.jobid)
        if action == JobAction.hold:
            fastr.log.debug('Holding {} until dependencies are met'.format(job.jobid))
            self.job_status[job.jobid] = job.status = JobState.hold
        elif action == JobAction.cancel:
            self.cancel_job(job)
        else:
            # Check if the job is ready to run or must be held
            rq_job = self.queue.enqueue(self.run_job,
                                        job.jobid,
                                        job.commandfile,
                                        job.stdoutfile,
                                        job.stderrfile,
                                        job_id=job.id,
                                        ttl=-1)
            self.rq_jobs[job.jobid] = rq_job

    def check_finished(self):
        while self.running:
            for job_id, rq_job in self.rq_jobs.items():
                # Check if job is finished
                if rq_job.is_finished or rq_job.is_failed:
                    job = self.job_dict[job_id]
                    if rq_job.is_failed:
                        job.status = JobState.execution_failed
                    else:
                        job.status = JobState.execution_done
                    self.job_finished(job)
                    del self.rq_jobs[job_id]

            time.sleep(1.0)

    @classmethod
    def run_job(cls, job_id, job_command, job_stdout, job_stderr):
        try:
            fastr.log.debug('Running job {}'.format(job_id))

            command = [sys.executable,
                       os.path.join(fastr.config.executionscript),
                       job_command]

            with open(job_stdout, 'w') as fh_stdout, open(job_stderr, 'w') as fh_stderr:
                proc = subprocess.Popen(command, stdout=fh_stdout, stderr=fh_stderr)
                proc.wait()
                fastr.log.debug('Subprocess finished')
            fastr.log.debug('Finished {}'.format(job_id))
        except Exception:
            exc_type, _, trace = sys.exc_info()
            exc_info = traceback.format_exc()
            fastr.log.error('Encountered exception ({}) during execution:\n{}'.format(exc_type.__name__, exc_info))
            raise



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module that defines the NetworkAnalyzer and holds the reference implementation.
"""
from abc import abstractmethod


class NetworkAnalyzer(object):
    """
    Base class for NetworkAnalyzers
    """
    @abstractmethod
    def analyze_network(self, network, chunk):
        """
        Analyze a chunk of a Network.

        :param network: Network corresponding with the chunk
        :param chunk: The chunk of the network to analyze
        """
        pass


class DefaultNetworkAnalyzer(NetworkAnalyzer):
    """
    Default implementation of the NetworkAnalyzer.
    """
    def analyze_network(self, network, chunk):
        """
        Analyze a chunk of a Network. Simply process the Nodes in the chunk
        sequentially.

        :param network: Network corresponding with the chunk
        :param chunk: The chunk of the network to analyze
        """
        _, nodes = chunk

        # Start should be able to execute as is
        executionlist = []

        # All the unassigned nodes
        candidates = [x for x in nodes]

        while len(candidates) > 0:
            processed = []
            for node in candidates:
                prerequisites = node.get_sourced_nodes()

                for required in prerequisites:
                    if required in candidates:
                        break
                else:
                    # All required nodes are handled, move the candidate to the back of the execution list
                    executionlist.append(node)
                    processed.append(node)

            for node in processed:
                candidates.remove(node)

        return executionlist



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module to maintain a network node.

Exported classes:

Node -- A class encapsulating a tool.
ConstantNode -- A node encapsulating an Output to set scalar values.
SourceNode -- A class providing a handle to a file.
"""
from abc import ABCMeta
import itertools
from collections import OrderedDict
import os
import re
import urlparse


#For dynamically importing/loading of files(MacroNodes)
import imp

import sympy
from sympy.core.symbol import Symbol

import fastr
from fastr.datatypes import DataType, Deferred
import fastr.exceptions as exceptions
from fastr.core.inputoutput import BaseInput, Input, BaseOutput, Output, AdvancedFlowOutput, SourceOutput
from fastr.core.samples import SampleId, SampleItem, SampleIndex, SampleValue, SamplePayload
from fastr.core.serializable import Serializable
from fastr.core.tool import Tool
from fastr.core.updateable import Updateable, UpdateableMeta
from fastr.data import url
from fastr.execution.job import Job, SinkJob, SourceJob, InlineJob, JobState


class InputDict(OrderedDict):
    """
    The container containing the Inputs of Node. Implements helper functions
    for the easy linking syntax.
    """

    # We know this class is not really for public interaction, however it has
    # an important function linking of nodes.
    # pylint: disable=too-few-public-methods

    def __setitem__(self, key, value, dict_setitem=dict.__setitem__):
        """
        Set an item in the input dictionary. The behaviour depends on the
        type of the value. For a
        :py:class:`BaseInput <fastr.core.inputoutput.BaseInput>`,
        the input will simply be added to the list of inputs. For a
        :py:class:`BaseOutput <fastr.core.inputoutput.BaseOutput>`,
        a link between the output and input will be created.

        :param str key: id of the input to assign/link
        :param value: either the input to add or the output to link
        :type value: :py:class:`BaseInput <fastr.core.inputoutput.BaseInput>`
                      or
                      :py:class:`BaseOutput <fastr.core.inputoutput.BaseOutput>`
        :param dict_setitem: the setitem function to use for the underlying
                              OrderedDict insert
        """
        if isinstance(value, Input):
            super(InputDict, self).__setitem__(key, value, dict_setitem=dict_setitem)
        elif isinstance(value, BaseOutput):
            if self[key].node.parent is not value.node.parent:
                message = 'Cannot create links between members of different Network'
                fastr.log.warning(message)

            network = value.node.parent
            if network is None:
                message = 'Cannot create links between non-network-attached Nodes'
                fastr.log.warning(message)
            else:
                fastr.log.debug('Linking {} to {}'.format(value.fullid, self[key].fullid))
                network.create_link(value, self[key])
        elif isinstance(value, (list, tuple, dict, OrderedDict)):
            # This is data for a ConstantNode, so create one and set it
            # First make sure the stepid of the new ConstantNode will match the stepid of the current Node
            inp = self[key]
            for k, i in inp.node.parent.stepids.items():
                if inp.node in i:
                    stepid = k
                    break
            else:
                stepid = None
            network = inp.node.parent
            new_id = 'const_{}_{}'.format(inp.node.id, inp.id)
            const_node = network.create_constant(datatype=inp.datatype, data=value, id_=new_id, stepid=stepid)
            network.create_link(const_node.output, self[key])

        else:
            message = 'Cannot add object of type {} to InputDict'.format(type(value).__name__)
            fastr.log.critical(message)
            raise exceptions.FastrTypeError(message)


class OutputDict(OrderedDict):
    """
    The container containing the Inputs of Node. Only checks if the inserted
    values are actually outputs.
    """

    # We know this class is not really for public interaction, however it we
    # have it to do type checking and consistency with the InputDict
    # pylint: disable=too-few-public-methods

    def __setitem__(self, key, value, dict_setitem=dict.__setitem__):
        """
        Set an output.

        :param str key: the of the item to set
        :param value: the output to set
        :type value: :py:class:`BaseOutput <fastr.core.inputoutput.BaseOutput>`
        :param dict_setitem: the setitem function to use for the underlying
                              OrderedDict insert
        """
        if isinstance(value, Output):
            super(OutputDict, self).__setitem__(key, value, dict_setitem=dict_setitem)
        else:
            message = 'Cannot add object of type {} to OutputDict'.format(type(value).__name__)
            fastr.log.warning(message)


class InputGroup(OrderedDict):
    """
    A class representing a group of inputs. Input groups allow the
    """
    __metaclass__ = UpdateableMeta
    __updatetriggers__ = ['__init__',
                          '__setitem__',
                          '__delitem__',
                          'clear',
                          'pop',
                          'popitem',
                          'setdefault',
                          'update']

    def __init__(self, parent, id_=None):
        """
        Create a new InputGroup representation

        :param parent: the parent node
        :type parent: :py:class:`Node <fastr.core.node.Node>`
        :param str id_: the id of the input group
        :raises FastrTypeError: if parent is not a Node
        """
        if not isinstance(parent, Node):
            raise exceptions.FastrTypeError('parent should be a Node')
        super(InputGroup, self).__init__()
        self._parent = parent
        self.id = id_
        self._size = ()
        self._dimnames = ()
        self._primary = None
        self.__updating__ = True

    def __getitem__(self, key):
        if isinstance(key, SampleIndex):
            index = key

            # Determine which input sample to use for Input
            indexmap = dict(zip(self.dimnames, index))
            data = {}

            nodegroups = self.parent.parent.nodegroups

            # Match dimensions if possible
            lookup = {v: dimname for dimname in self.dimnames for value in nodegroups.values() if dimname in value for v in value}
            lookup.update({x: x for x in self.dimnames})

            for key, value in lookup.items():
                lookup[key] = indexmap[value]

            for id_, input_ in self.items():
                source_index = self.find_source_index(target_size=self.size,
                                                      target_dimnames=self.dimnames,
                                                      source_size=input_.size,
                                                      source_dimnames=input_.dimnames,
                                                      target_index=index)

                # Get data from Input
                if source_index is not None:
                    data[id_] = input_[source_index]
                else:
                    data[id_] = SampleItem(index, '__EMPTY__', [], set())

            # Aggregate data for input group
            sample_id = data[self.primary.id].id
            hold_jobs = set.union(*[val.jobs for val in data.values()])

            return SamplePayload(index, sample_id, data, hold_jobs)
        else:
            return super(InputGroup, self).__getitem__(key)

    def __setitem__(self, key, value, dict_setitem=dict.__setitem__):
        """
        Assign an input to this input group.

        :param str key: id of the input
        :param value: the input to assign
        :type value: :py:class:`Input <fastr.core.inputoutput.Input>`
        :raises FastrTypeError: if value of valid type
        """
        if not isinstance(value, Input):
            message = 'Cannot assign a non-Input to an InputGroup'
            fastr.log.error(message)
            raise exceptions.FastrTypeError(message)

        if self._parent is not None and value.node is not self._parent:
            message = 'Input has a different parent Node than the InputGroup, this is not a valid assignment!'
            fastr.log.error(message)
            raise exceptions.FastrParentMismatchError(message)

        super(InputGroup, self).__setitem__(key, value)

    @property
    def parent(self):
        """
        The parent node of this InputGroup
        """
        return self._parent

    @property
    def size(self):
        """
        The sample size of this InputGroup
        """
        return self._size

    @property
    def dimnames(self):
        """
        The names of the dimensions in this InputGroup
        """
        return self._dimnames

    @property
    def primary(self):
        """
        The primary Input in this InputGroup. The primary Input is the Input
        that defines the size of this InputGroup. In case of ties it will be
        the first in the tool definition.
        """
        return self._primary

    @property
    def empty(self):
        """
        Bool indicating that this InputGroup is empty (has no data connected)
        """
        return self.size is None or len([x for x in self.size if x != 0]) == 0

    @classmethod
    def find_source_index(cls, target_size, target_dimnames, source_size, source_dimnames, target_index):
        # Determine which input sample to use for Input

        if source_size == target_size:
            source_index = target_index
        elif source_size == (1,):
            source_index = SampleIndex(0)
        elif all(x == 0 for x in source_size):
            source_index = None
        else:
            source_index = cls.solve_broadcast(target_size=target_size,
                                               target_dimnames=target_dimnames,
                                               source_size=source_size,
                                               source_dimnames=source_dimnames,
                                               target_index=target_index)

        return source_index

    @staticmethod
    def _get_lookup(target_dimnames, nodegroups=None):
        if nodegroups is None:
            nodegroups = fastr.current_network.nodegroups

        lookup = {v: dimname for dimname in target_dimnames for value in nodegroups.values() if dimname in value for v in value}
        lookup.update({x: x for x in target_dimnames})

        return lookup

    @classmethod
    def solve_broadcast(cls, target_size, target_dimnames, source_size, source_dimnames, target_index, nodegroups=None):
        indexmap = dict(zip(target_dimnames, target_index))
        sizemap = dict(zip(target_dimnames, target_size))
        lookup = cls._get_lookup(target_dimnames, nodegroups)

        if all(x in lookup for x in source_dimnames):
            matched_dims = [lookup[x] for x in source_dimnames]
            source_index = SampleIndex(indexmap[x] for x in matched_dims)
            estimated_source_size = tuple(sizemap[x] for x in matched_dims)
        else:
            raise exceptions.FastrSizeMismatchError('Cannot broadcast, not all dimnames can be matched! (source dimnames {}, lookup {}'.format(source_dimnames, lookup))

        if source_size != estimated_source_size:
            raise exceptions.FastrSizeMismatchError('The estimated size after broadcast matching is incorrect! (estimated {}, actual {})'.format(estimated_source_size, source_size))

        return source_index

    @property
    def iterinputvalues(self):
        """
        Iterate over the item in this InputGroup

        :returns: iterator yielding :py:class:`SampleItems <fastr.core.sampeidlist.SampleItem>`
        """
        for index, _, _, _ in self.primary.iteritems():
            yield self[index]

    def __updatefunc__(self):
        """
        Update the InputGroup. Triggers when a change is made to the content
        of the InputGroup. Automatically recalculates the size, primary Input
        etc.
        """
        sizes = [x.size for x in self.values()]
        dimnames = [x.dimnames for x in self.values()]

        unique_sizes = set(sizes) - {(0,), (1,), (), None}

        if len(unique_sizes) > 1:
            if not any(all(not isinstance(x, Symbol) for x in size) for size in unique_sizes):
                # All entries in unique sizes are symbols, we cannot really
                # know what will be the size. Assume for now that the first
                # Input with symbolic input will be the primary
                max_dims = max(len(x.dimnames) for x in self.values())
                for input_ in self.values():
                    if input_.size in unique_sizes and len(input_.dimnames) == max_dims:
                        self._size = input_.size
                        self._primary = input_
                        self._dimnames = input_.dimnames
                        break

            # Check if we can match via dimnames
            elif all(len(x) == len(set(x)) for x in dimnames):
                longest_dimname, longest_size = max(zip(dimnames, sizes), key=lambda x: len(x[1]))

                if all(all(x in longest_dimname for x in y) for y in dimnames):
                    self._size = longest_size
                    self._primary = self.values()[sizes.index(self._size)]
                    self._dimnames = longest_dimname
                else:
                    nodegroups = self.parent.parent.nodegroups
                    lookup = {v: dimname for dimname in longest_dimname for value in nodegroups.values() if dimname in value for v in value}
                    lookup.update({x: x for x in longest_dimname})

                    if all(x in lookup for y in dimnames for x in y):
                        self._size = longest_size
                        self._primary = self.values()[sizes.index(self._size)]
                        self._dimnames = longest_dimname
                    else:
                        message = 'Not all dimnames ({}) are present in the highest-dimensional input ({})'.format(
                            [x for y in dimnames for x in y],
                            lookup.keys(),
                        )
                        fastr.log.warning(message)

                        self._size = longest_size
                        self._primary = self.values()[sizes.index(self._size)]
                        self._dimnames = longest_dimname
            else:
                message = 'One or more inputs have non-unique dimnames ({})'.format(dimnames)
                fastr.log.error(message)
                raise exceptions.FastrValueError(message)

        elif len(unique_sizes) == 1:
            self._size = unique_sizes.pop()
            self._primary = self.values()[sizes.index(self._size)]
            self._dimnames = self._primary.dimnames
        elif (1,) in sizes:
            self._size = (1,)
            self._primary = self.values()[sizes.index(self._size)]
            self._dimnames = self._primary.dimnames
        else:
            self._size = ()
            self._primary = None
            self._dimnames = ()


class DefaultInputGroupCombiner(object):
    def __init__(self, input_groups):
        self.input_groups = input_groups
        self.update()

    def merge(self, list_of_items):
        """
        Given a list of items for each input group, it returns the combined
        list of items.

        :param list list_of_items: items to combine
        :return: combined list
        """
        return [x for item in list_of_items for x in item]

    def unmerge(self, item):
        """
        Given a item it will recreate the seperate items, basically this is the
        inverse operation of merge. However, this create an OrderedDict so that
        specific input groups can be easily retrieved. To get a round trip, the
        values of the OrderedDict should be taken::

            >>> list_of_items = combiner.unmerge(item)
            >>> item = combiner.merge(list_of_items.values())

        :param list item: the item to unmerge
        :return: items
        :rtype: OrderedDict
        """
        result = OrderedDict()
        for target in self.input_groups.values():
            mask = [True if ig.id == target.id else False for ig in self.input_groups.values() for _ in ig.size]
            result[target.id] = tuple(k for k, m in zip(item, mask) if m)

        return result

    @property
    def dimnames(self):
        dimnames = tuple(self.merge([ig.dimnames for ig in self.input_groups.values()]))
        return dimnames

    @property
    def outputsize(self):
        return tuple(self.merge([ig.size for ig in self.input_groups.values()]))

    def merge_sample_id(self, list_of_sample_ids):
        return SampleId(self.merge(list_of_sample_ids))

    def merge_sample_index(self, list_of_sample_indexes):
        return SampleIndex(self.merge(list_of_sample_indexes))

    def merge_sample_data(self, list_of_sample_data):
        return {k: v if v is not None and len(v) > 0 else None for data in list_of_sample_data for k, v in data.items()}

    def merge_sample_jobs(self, list_of_sample_jobs):
        return set.union(*list_of_sample_jobs)

    def merge_payloads(self, sample_payloads):
        # Determine the sample index
        sample_index = self.merge_sample_index([x.index for x in sample_payloads])

        # Create sampleid
        sample_id = self.merge_sample_id([x.id for x in sample_payloads])

        # Extract jobdata and combine in single dict
        job_data = self.merge_sample_data(x.data for x in sample_payloads)

        # Create superset of all job dependencies
        job_depends = self.merge_sample_jobs(x.jobs for x in sample_payloads)

        return sample_index, sample_id, job_data, job_depends

    def iter_input_groups(self):
        for sample_payloads in itertools.product(*[ig.iterinputvalues for ig in self.input_groups.values()]):
            fastr.log.debug('sample_payload: {}'.format(sample_payloads))
            fastr.log.debug('sample payload data: {}'.format([x.data for x in sample_payloads]))
            yield self.merge_payloads(sample_payloads)

    def __iter__(self):
        return self.iter_input_groups()

    def update(self):
        pass


class MergingInputGroupCombiner(DefaultInputGroupCombiner):
    def __init__(self, input_groups, merge_dimension):
        self.merge_dimensions = merge_dimension
        self._merges = None
        self._merge_sizes = None
        self._masks = None
        super(MergingInputGroupCombiner, self).__init__(input_groups)

    def update(self):
        dimnames = [x.dimnames for x in self.input_groups.values()]
        sizes = [x.size for x in self.input_groups.values()]

        # Validate the sample dimensions and sizes are consistent
        unique_dimnames = tuple(set(x for dimname in dimnames for x in dimname))
        unique_sizes = {x: set() for x in unique_dimnames}

        for size, dimname in zip(sizes, dimnames):
            for size_element, dimname_element in zip(size, dimname):
                unique_sizes[dimname_element].add(size_element)
        if not all(len(x) == 1 for x in unique_sizes.values()):
            raise exceptions.FastrValueError('The dimension have incosistent sizes: {}'.format(unique_sizes))
        unique_sizes = {k: v.pop() for k, v in unique_sizes.items()}

        # Check how many merges to perform
        if self.merge_dimensions == 'all':
            counts = {name: [sum(x == name for x in dimname) for dimname in dimnames] for name in unique_dimnames}
            merges = {name: min(value) for name, value in counts.items()}
            merges = tuple(key for key, value in merges.items() for x in range(value))
        else:
            merges = tuple(self.merge_dimensions)

        self._merges = merges

        # Make a merge mask for the dimnames
        masks = []
        for dimname in dimnames:
            temp = []
            temp_merges = list(merges)
            for name in dimname:
                if name in temp_merges:
                    index = temp_merges.index(name)
                    temp.append(index)
                    temp_merges[index] = None
                else:
                    temp.append(slice(unique_sizes[name]))
            masks.append(temp)

        self._masks = masks

        # Do the actual merging
        self._merge_sizes = tuple(unique_sizes[x] for x in merges)

    def merge(self, list_of_items):
        new_item = [x for mask, item in zip(self._masks, list_of_items) for m, x in zip(mask, item) if isinstance(m, slice)]
        new_item.extend(x for x, m in zip(list_of_items[0], self._masks[0]) if not isinstance(m, slice))
        return new_item

    def unmerge(self, item):
        index = 0
        result = OrderedDict()
        nr_slices = sum(isinstance(x, slice) for y in self._masks for x in y)
        for input_group, mask in zip(self.input_groups, self._masks):
            original_item = []
            for m in mask:
                if isinstance(m, slice):
                    original_item.append(item[index])
                    index += 1
                else:
                    original_item.append(item[m + nr_slices])
            result[input_group] = tuple(original_item)

        return result

    def iter_input_groups(self):
        # Create iterator for the merged part of the iteration
        if len(self._merges) > 0:
            fixed_indexes = itertools.product(*[xrange(x) for x in self._merge_sizes])
        else:
            fixed_indexes = [()]

        # Loop over the fixed (merged) indices and then over the remainder
        for fixed_index in fixed_indexes:
            iterators = [itertools.product(*[xrange(m.stop) if isinstance(m, slice) else [fixed_index[m]] for m in mask]) for mask in self._masks]

            for indexes in itertools.product(*iterators):
                # Retrieve the samples from all input groups
                samples = []

                for index, input_group in zip(indexes, self.input_groups.values()):
                    try:
                        samples.append(input_group[SampleIndex(index)])
                    except exceptions.FastrIndexNonexistent:
                        samples.append(None)

                # Merge all sample payloads
                if all(x is not None for x in samples):
                    yield self.merge_payloads(samples)
                else:
                    fastr.log.info('Skipping {} because it is non-existent due to sparsity!'.format(indexes))


class Node(Updateable, Serializable):
    """
    The class encapsulating a node in the network. The node is responsible
    for setting and checking inputs and outputs based on the description
    provided by a tool instance.
    """

    __metaclass__ = ABCMeta
    __dataschemafile__ = 'Node.schema.json'

    _InputType = Input
    _OutputType = Output
    _JobType = Job

    def __init__(self, tool, id_=None, parent=None, cores=None, memory=None, walltime=None):
        """
        Instantiate a node.

        :param tool: The tool to base the node on
        :type tool: :py:class:`Tool <fastr.core.tool.Tool>`
        :param str id_: the id of the node
        :param parent: the parent network of the node
        :param int cores: number of cores required for executing this Node
        :param str memory: amount of memory required in the form \\d+[mMgG]
                           where M is for megabyte and G for gigabyte
        :param str walltime: amount of time required in second or in the form
                             HOURS:MINUTES:SECOND
        :type parent: :py:class:`Network <fastr.core.network.Network>`
        :return: the newly created Node
        """
        super(Node, self).__init__()

        if isinstance(tool, Tool):
            self._tool = tool
        elif isinstance(tool, (str, tuple)):
            if tool in fastr.toollist:
                self._tool = fastr.toollist[tool]
            else:
                message = ('Specified tool ({}) is not in the toollist: {}. '
                           'Check the config (fastr/resources/fastr.config)').format(tool,
                                                                                     fastr.toollist.todict().keys())
                fastr.log.critical(message)
                raise exceptions.FastrToolUnknownError(message)
        elif tool is None:
            self._tool = None
        else:
            message = 'tool should either be a string or a Tool.'
            fastr.log.critical(message)
            raise exceptions.FastrTypeError(message)

        # Don't set parent here, as not info needed for registration is there yet
        self._parent = None
        if parent is not None:
            parent = parent
        elif fastr.current_network is not None:
            parent = fastr.current_network
        else:
            message = 'Both parent argument and fastr.current_network are None, need a parent Network to function!'
            raise exceptions.FastrValueError(message)

        node_number = parent.node_number
        parent.node_number += 1

        if id_ is None:
            # Node ID is a simple $name_$counter format, making sure nodes can
            # not be named the same

            #: The Node id s a unique string identifying the Node
            id_ = 'node_{}_{}'.format(self.name, node_number)

        parent.check_id(id_)
        self._id = id_

        #: The parent is the Network this Node is part of
        self.parent = parent

        #: A list of inputs of this Node
        self.inputs = InputDict()

        #: A list of outputs of this Node
        self.outputs = OutputDict()

        # Create all inputs and outputs, if the class is set in the Tool file,
        # respect that, otherwise use the class default.
        if self._tool is not None:
            for name, input_ in self._tool.inputs.items():
                self.inputs[name] = self._InputType(self, input_)
            for name, output in self._tool.outputs.items():
                self.outputs[name] = self._OutputType(self, output)

        self._input_groups = OrderedDict()
        self.jobs = None

        # Set the job requirements
        self._required_cores = None
        self._required_memory = None
        self._required_time = None
        self.required_cores = cores
        self.required_memory = memory
        self.required_time = walltime

        # Set the flow control
        self._merge_dimensions = None
        self._input_group_combiner = None
        self.merge_dimensions = 'none'

        # Update Inputs and self (which calls Outputs)
        self.update()

    def __repr__(self):
        """
        Get a string representation for the Node

        :return: the string representation
        :rtype: str
        """
        if self._tool is not None:
            toolinfo = '(tool: {tool.id} v{tool.version!s})'.format(tool=self._tool)
        else:
            toolinfo = ''
        return_list = ['{} {} {}'.format(type(self).__name__, self.id, toolinfo)]

        # The "+ [8]" guarantees a minimum of 8 width and avoids empty lists
        width_input_keys = max([len(x) for x in self.inputs.keys()] + [8])
        width_input_types = max([len(x.datatype.id) for x in self.inputs.values()] + [8]) + 2
        width_output_keys = max([len(x) for x in self.outputs.keys()] + [8])
        width_output_types = max([len(x.datatype.id) for x in self.outputs.values()] + [8]) + 2

        return_list.append('{:^{}}  | {:^{}}'.format('Inputs', width_input_types + width_input_keys + 1,
                                                     'Outputs', width_output_types + width_output_keys + 1))
        return_list.append('-' * (width_input_keys + width_input_types + width_output_keys + width_output_types + 7))
        for (input_key, input_, output_key, output) in itertools.izip_longest(self.inputs.keys(), self.inputs.values(), self.outputs.keys(), self.outputs.values()):
            if input_ is None:
                input_id = ''
                input_type = ''
            else:
                input_id = input_key
                input_type = '({})'.format(input_.datatype.id)

            if output is None:
                output_id = ''
                output_type = ''
            else:
                output_id = output_key
                output_type = '({})'.format(output.datatype.id)

            return_list.append('{:{}} {:{}}  |  {:{}} {:{}}'.format(input_id, width_input_keys,
                                                                    input_type, width_input_types,
                                                                    output_id, width_output_keys,
                                                                    output_type, width_output_types))

        return '\n'.join(return_list)

    def __str__(self):
        """
        Get a string version for the Node

        :return: the string version
        :rtype: str
        """
        return "<{}: {}>".format(type(self).__name__, self.id)

    def __eq__(self, other):
        """Compare two Node instances with each other. This function ignores
        the parent and update status, but tests rest of the dict for equality.
        equality

        :param other: the other instances to compare to
        :type other: Node
        :returns: True if equal, False otherwise
        """
        if not isinstance(other, Node):
            return NotImplemented

        dict_self = {k: v for k, v in self.__dict__.items()}
        del dict_self['_parent']
        del dict_self['_status']
        del dict_self['_input_groups']
        del dict_self['jobs']
        del dict_self['_input_group_combiner']

        dict_other = {k: v for k, v in other.__dict__.items()}
        del dict_other['_parent']
        del dict_other['_status']
        del dict_other['_input_groups']
        del dict_other['jobs']
        del dict_other['_input_group_combiner']

        return dict_self == dict_other

    def __getstate__(self):
        """
        Retrieve the state of the Node

        :return: the state of the object
        :rtype dict:
        """
        state = super(Node, self).__getstate__()

        # Make id prettier in the result
        state['id'] = self.id

        # Add the class of the Node in question
        state['class'] = type(self).__name__

        # Get all input and output
        state['inputs'] = [x.__getstate__() for x in self.inputs.values()]
        state['outputs'] = [x.__getstate__() for x in self.outputs.values()]

        if self._tool is not None:
            state['tool'] = [self._tool.ns_id, str(self._tool.command['version'])]
        else:
            state['tool'] = None

        # Add resource requirements
        state['required_cores'] = self._required_cores
        state['required_memory'] = self._required_memory
        state['required_time'] = self._required_time
        state['merge_dimensions'] = self._merge_dimensions

        return state

    def __setstate__(self, state):
        """
        Set the state of the Node by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        # Make sure the Node classes are aligned (and warn if not so)
        if 'class' in state and state['class'] != type(self).__name__:
            fastr.log.warning('Attempting to set the state of a {} to {}'.format(
                state['class'],
                type(self).__name__
            ))

        self.jobs = None

        if not hasattr(self, '_input_groups'):
            self._input_groups = OrderedDict()

        if 'id' in state:
            self._id = state['id']

        if 'parent' in state:
            parent = state['parent']
            del state['parent']
        else:
            parent = None

        if state['tool'] is not None:
            self._tool = fastr.toollist[tuple(state['tool'])]
        else:
            self._tool = None

        # Create Input/Output objects
        inputlist = []
        for input_ in state['inputs']:
            if 'node' in input_:
                # Check if the expected Node id matches our current id
                if input_['node'] != state['id']:
                    raise exceptions.FastrParentMismatchError('This Input has different parent node!')
                del input_['node']

            # It can happen that this has been done by a subclass, check first
            if not isinstance(input_, BaseInput):
                description = self.tool.inputs[input_['id']]
                inputobj = self._InputType(self, description)
                inputobj._node = self
                inputobj.__setstate__(input_)
            else:
                inputobj = input_
            inputlist.append((inputobj.id, inputobj))

        outputlist = []
        for output in state['outputs']:
            if '_node' in output:
                # Check if the expected Node id matches our current id
                if output['_node'] != state['_id']:
                    raise exceptions.FastrParentMismatchError('This Input has different parent node!')
                del output['_node']

            # It can happen that this has been done by a subclass, check first
            if not isinstance(output, Output):
                description = self.tool.outputs[output['id']]
                outputobj = self._OutputType(self, description)
                outputobj.__setstate__(output)
            else:
                outputobj = output
            outputlist.append((outputobj.id, outputobj))

        self.inputs = InputDict(inputlist)
        self.outputs = OutputDict(outputlist)

        super(Node, self).__setstate__(state)

        self._parent = None
        if parent is not None:
            self.parent = parent
        elif fastr.current_network is not None:
            self.parent = fastr.current_network
        else:
            message = 'Both parent argument and fastr.current_network are None, need a parent Network to function!'
            raise exceptions.FastrValueError(message)

        self._required_cores = state['required_cores']
        self._required_memory = state['required_memory']
        self._required_time = state['required_time']

        self.merge_dimensions = state['merge_dimensions']

        self.update()

    @property
    def merge_dimensions(self):
        return self._merge_dimensions

    @merge_dimensions.setter
    def merge_dimensions(self, value):
        if isinstance(value, (str, unicode)):
            options = ['all', 'none']
            if value not in options:
                raise exceptions.FastrValueError('Invalid option {} given (valid options: {})'.format(value, options))
            self._merge_dimensions = value
            if value == 'none':
                self._input_group_combiner = DefaultInputGroupCombiner(self.inputgroups)
            elif value == 'all':
                self._input_group_combiner = MergingInputGroupCombiner(self.inputgroups, value)
        else:
            self._merge_dimensions = value
            self._input_group_combiner = MergingInputGroupCombiner(self.inputgroups, tuple(value))

    @classmethod
    def createobj(cls, state, network=None):
        if 'parent' not in state or not isinstance(state['parent'], fastr.Network):
            if network is not None:
                fastr.log.debug('Setting network to: {}'.format(network))
                state['parent'] = network
            else:
                fastr.log.debug('No network given for de-serialization')
        else:
            fastr.log.debug('Parent is already defined as: {}'.format(network))

        state = dict(state)

        return super(Node, cls).createobj(state, network)

    @property
    def blocking(self):
        """
        Indicate that the results of this Node cannot be determined without first executing the Node, causing a
        blockage in the creation of jobs. A blocking Nodes causes the Chunk borders.
        """
        for output in self.outputs.values():
            if output.blocking:
                fastr.log.info('Blocking because Output {} has cardinality {}'.format(output.fullid,
                                                                                      output.cardinality()))
                return True
        return False

    @property
    def dimnames(self):
        """
        Names of the dimensions in the Node output. These will be reflected
        in the SampleIdList of this Node.
        """
        if hasattr(self, '_dimnames') and self._dimnames is not None:
            return self._dimnames
        else:
            return self._input_group_combiner.dimnames

    @dimnames.setter
    def dimnames(self, value):
        if isinstance(value, str):
            value = value,

        if not isinstance(value, tuple) or not all(isinstance(x, str) for x in value):
            raise exceptions.FastrTypeError('Dimnames has to be a tuple of str!')

        fastr.log.warning('You are overriding the dimnames of a Node, beware this is possible but not encouraged and can lead to strange results!')
        self._dimnames = value

    @dimnames.deleter
    def dimnames(self):
        del self._dimnames

    @property
    def fullid(self):
        """
        The full defining ID for the Node
        """
        return '{}/nodelist/{}'.format(self.parent.fullid, self.id)

    @property
    def inputgroups(self):
        """
        A list of inputgroups for this Node. An input group is InputGroup
         object filled according to the Node

        """
        return self._input_groups

    @property
    def outputsize(self):
        """
        Size of the outputs in this Node
        """
        # Get sizes of all input groups
        return self._input_group_combiner.outputsize

    @property
    def id(self):
        """
        The id of the Node
        """
        return self._id

    @property
    def listeners(self):
        """
        All the listeners requesting output of this node, this means the
        listeners of all Outputs and SubOutputs
        """
        return {l for output in self.outputs.values() for l in output.listeners}

    @property
    def name(self):
        """
        Name of the Tool the Node was based on. In case a Toolless Node was
        used the class name is given.
        """
        if hasattr(self, '_tool') and isinstance(self._tool, Tool):
            return self._tool.id
        else:
            return self.__class__.__name__

    @property
    def nodegroup(self):
        for name, group in self.parent.nodegroups.items():
            if self.id in group:
                return name
        return None

    @property
    def parent(self):
        """
        The parent network of this node.
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        """
        The parent network of this node. (setter)
        """
        if self._parent is value:
            return  # Setting to same value doesn't do anything

        if self._parent is not None:
            fastr.log.warning('EXCEPTION|FastrAttributeError| Cannot reset attribute once set {} --> {}'.format(self._parent.id, value.id))
            #raise exceptions.FastrAttributeError('Cannot reset attribute once set')


        self._parent = value
        self._parent.add_node(self)

    @property
    def required_cores(self):
        """
        Number of cores required for the execution of this Node
        """
        return self._required_cores

    @required_cores.setter
    def required_cores(self, value):
        """
        Number of cores required for the execution of this Node (setter)
        """
        if value is None:
            self._required_cores = value
        else:
            if not isinstance(value, int):
                raise TypeError('Required number of cores should be an integer or None')

            if value < 1:
                raise ValueError('Required number of cores should be above zero ({} < 1)'.format(value))

            self._required_cores = value

    @property
    def required_memory(self):
        """
        Amount of memory required for the execution of this Node. Follows
        the format \\d+[mMgG] so 500M or 4g would be valid ways to specify
        500 megabytes or 4 gigabyte of memory.
        """
        return self._required_memory

    @required_memory.setter
    def required_memory(self, value):
        """
        Amount of memory required for the execution of this Node. Follows
        the format \\d+[mMgG] so 500M or 4g would be valid ways to specify
        500 megabytes or 4 gigabyte of memory. (setter)
        """
        if value is None:
            self._required_memory = value
        else:
            if not isinstance(value, str):
                raise TypeError('Required memory should be a str or None')

            if re.match(r'^\d+[mMgG]$', value) is None:
                raise ValueError('Required memory should be in the form \\d+[mMgG] (found {})'.format(value))

            self._required_memory = value

    @property
    def required_time(self):
        """
        Amount of time required for the execution of this Node. Follows the
        format of a number of second or H:M:S, with H the number of hours,
        M the number of minutes and S the number of seconds.
        """
        return self._required_time

    @required_time.setter
    def required_time(self, value):
        """
        Amount of time required for the execution of this Node. Follows the
        format of a number of second or H:M:S, with H the number of hours,
        M the number of minutes and S the number of seconds. (setter)
        """
        if value is None:
            self._required_time = value
        else:
            if not isinstance(value, str):
                raise TypeError('Required time should be a str or None')

            if re.match(r'^(\d*:\d*:\d*|\d+)$', value) is None:
                raise ValueError('Required time should be in the form HH:MM:SS or MM:SS (found {})'.format(value))

            self._required_time = value

    @property
    def status(self):
        return self._status

    @property
    def tool(self):
        return self._tool

    def execute(self):
        """
        Execute the node and create the jobs that need to run

        :return: list of jobs to run
        :rtype: list of :py:class:`Jobs <fastr.execution.job.Job>`
        """
        self.update(False, False)

        # Make sure a Node is valid
        if not self.valid:
            message = 'Node {} is not valid'.format(self.fullid)
            fastr.log.error(message)
            fastr.log.error('Messages:\n{}'.format('\n'.join(self.messages)))
            raise exceptions.FastrNodeNotValidError(message)
        inputgroups = self.inputgroups

        # Define output size and dimension names
        ig_masters = [ig.primary.id for ig in inputgroups.values()]
        fastr.log.debug('size: {} dimnames: {} masters: {}'.format(self.outputsize, self.dimnames, ig_masters))

        # Prepare the output of the Node
        fastr.log.debug('Preparing {} with size {} and dimnames {}'.format(self.fullid, self.outputsize, self.dimnames))
        self.prepare()

        # Iterate over all combinations of inputgroups to create sets of data
        job_list = []

        fastr.log.debug('InputGroups: {}'.format(inputgroups.values()))
        fastr.log.debug('Inputs: {}'.format([x for ig in inputgroups.values() for x in ig.values()]))
        fastr.log.debug('Sources: {}'.format([x.source for ig in inputgroups.values() for x in ig.values()]))

        for sample_index, sample_id, job_data, job_depends in self._input_group_combiner:
            fastr.log.debug('----- START -----')
            fastr.log.debug('INDEX: {}'.format(sample_index))
            fastr.log.debug('SAMPLE_ID: {} {}'.format(repr(sample_id), sample_id))
            fastr.log.debug('JOBDATA: {}'.format(job_data))
            fastr.log.debug('JOBDEPS: {}'.format(job_depends))
            fastr.log.debug('------ END ------')

            job_list.append(self.create_job(sample_id, sample_index, job_data, job_depends))

        fastr.log.debug('joblist: {}'.format(job_list))
        fastr.log.debug('===== END execute_node =====')
        return job_list

    def set_result(self, job):
        """
        Incorporate result of a job into the Node.

        :param Type job: job of which the result to store
        """
        sample_id, sample_index = job.sample_id, job.sample_index
        # Replace following code by node.set_data(job.output_data) ? or something like it?
        for output in self.outputs.values():
            if output.id not in job.output_data:
                if len(output.listeners) > 0:
                    error_message = 'Could not find required data for {} in {}!'.format(output.fullid, job.output_data)
                    fastr.log.error(error_message)
                    raise exceptions.FastrValueError(error_message)

                continue

            output_data = job.output_data[output.id]
            fastr.log.info('Setting data for blocking node: {} sample: {}'.format(output.fullid, sample_id))

            output_url = '{}/{}/{}'.format(self.parent.tmpurl, self.id, sample_id)
            parsed_url = urlparse.urlparse(output_url)

            output_values = [urlparse.urlunparse(('val',
                                                  parsed_url.netloc,
                                                  url.join(parsed_url.path, '__fastr_result__.pickle.gz'),
                                                  parsed_url.params,
                                                  'outputname={}&nr={}'.format(output.id, c),
                                                  '')) for c in range(len(output_data))]

            output_values = tuple(Deferred(x) for x in output_values)
            fastr.log.debug('Setting collected for {} sample_id {} data: {}'.format(output.fullid,
                                                                                    sample_id,
                                                                                    output_values))
            output[sample_id, sample_index] = SampleItem(sample_index,
                                                         sample_id,
                                                         OrderedDict({0: tuple(output_values)}),
                                                         {job})
            self.jobs[sample_id] = job

    def create_job(self, sample_id, sample_index, job_data, job_dependencies, jobid=None, outputurl=None, **kwargs):
        """
        Create a job based on the sample id, job data and job dependencies.

        :param sample_id: the id of the corresponding sample
        :type sample_id: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :param dict job_data: dictionary containing all input data for the job
        :param job_dependencies: other jobs that need to finish before this job can run
        :return: the created job
        :rtype: :py:class:`Job <fastr.execution.job.Job>`
        """
        fastr.log.info('Creating job for node {} sample id {!r}, index {!r}'.format(self.fullid, sample_id, sample_index))
        fastr.log.debug('Creating job for sample {} with data {}'.format(sample_id, job_data))

        # Determine subfolder name in tmp mount
        if outputurl is None:
            outputurl = '{}/{}/{}'.format(self.parent.tmpurl, self.id, sample_id)

        # Determine absolute location of output dir and create directory
        outputdir = url.get_path_from_url(outputurl)

        fastr.log.debug('Sample ID: {}'.format(sample_id))
        fastr.log.debug('VFS Output dir: {}'.format(outputurl))

        # Remove output directory if there is old stuff present
        if not os.path.exists(outputdir):
            os.makedirs(outputdir)

        if not os.path.exists(outputdir):
            fastr.log.critical('Could not create output directory {}!'.format(outputdir))

        # Get the arguments
        input_arguments, output_arguments = self._wrap_arguments(job_data, sample_id, sample_index)

        preferred_types = {output.id: output.preferred_types for output in self.outputs.values()}

        if jobid is None:
            jobid = '{}___{}___{}'.format(self.parent.id, self.id, sample_id)
        fastr.log.debug('Job tempdir set to: {}'.format(outputurl))
        job = self._JobType(jobid, self._tool.ns_id, self._tool.command['version'], self.fullid, sample_id, sample_index, input_arguments, output_arguments, outputurl, job_dependencies, cores=self.required_cores, memory=self.required_memory, walltime=self.required_time, status_callback=self.parent.job_status_callback, preferred_types=preferred_types, **kwargs)
        self.jobs[sample_id] = job

        # Check which outputs are required or connected and set them
        if not self.blocking:
            for output in self.outputs.itervalues():
                fastr.log.debug('Preparing output {}'.format(output.id))
                if output.required or len(output.listeners) > 0:
                    fastr.log.debug('Setting {}'.format(output.id))
                    value = []
                    fastr.log.debug('Cardinality request: spec: {}, job_data: {}, and index: {}'.format(output.cardinality_spec, job_data, sample_index))
                    cardinality = output.cardinality(sample_index, job_data)
                    fastr.log.debug('Cardinality for {} is {}'.format(output.id, cardinality))
                    if not isinstance(cardinality, int):
                        message = 'For execution cardinality should be an int, for output ' \
                                  '{} we found {} (type {})'.format(output.id,
                                                                    cardinality,
                                                                    type(cardinality).__name__)
                        fastr.log.critical(message)
                        raise exceptions.FastrTypeError(message)

                    for cardinality_nr in range(0, cardinality):
                        parsed_url = urlparse.urlparse(outputurl)
                        value.append(urlparse.urlunparse(('val',
                                                          parsed_url.netloc,
                                                          url.join(parsed_url.path, '__fastr_result__.pickle.gz'),
                                                          parsed_url.params,
                                                          'outputname={}&nr={}'.format(output.id, cardinality_nr),
                                                          '')))

                    # Cast if we can, but otherwise leave and attempt it later
                    value = tuple(Deferred(x) for x in value)

                    output[sample_id] = SampleItem(sample_index,
                                                   sample_id,
                                                   {0: value},
                                                   {job})
        else:
            fastr.log.debug('Cannot determine blocking node output a priori! Needs to be collected afterwards!')

        return job

    def _wrap_arguments(self, job_data, sample_id, sample_index):
        """
        Wrap arguments into a list of tuples that the execution script can parse

        :param dict job_data: dictionary containing all input data for the job
        :param sample_id: the id of the corresponding sample
        :type sample_id: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the wrapped arguments in a tuple with the form ``(inputs, outputs)``
        """
        arguments = ({}, {})  # format is (input_args, output_args)
        for key, input_ in self.inputs.items():
            # Skip inputs that have no data
            if job_data[key] is None:
                if input_.default is not None:
                    # Insert the default data if present
                    job_data[key] = [input_.datatype(input_.default)]
                elif input_.required:
                    fastr.log.debug('Job data: {}'.format(job_data))
                    raise exceptions.FastrValueError('Node "{}" is missing data for required Input "{}"'.format(self.id, input_.id))
                else:
                    continue

            arguments[0][key] = job_data[key]

        for key, output in self.outputs.items():
            if not output.automatic:
                if output.required or len(output.listeners) > 0:
                    cardinality = output.cardinality(job_data)
                    arguments[1][key] = {'id': key,
                                         'cardinality': cardinality if cardinality is not None else output._output_cardinality,
                                         'datatype': output.resulting_datatype.id}

        return arguments

    def get_sourced_nodes(self):
        """
        A list of all Nodes connected as sources to this Node

        :return: list of all nodes that are connected to an input of this node
        """
        return list(set([n for input_ in self.inputs.values() for n in input_.get_sourced_nodes()]))

    def find_source_index(self, target_index, target, source):
        # If there are multiple inputgroups, select only part of index from
        # the inputgroup which source belongs to
        if len(self.inputgroups) > 1:
            inputgroups = self.inputgroups
            mask = [True if ig.id == source.input_group else False for ig in inputgroups.values() for _ in ig.size]
            target_index = tuple(k for k, m in zip(target_index, mask) if m)

        # Delegate to InputGroup to check mixing within InputGroup
        return self.inputgroups[source.input_group].find_source_index(target_size=target.size,
                                                                      target_dimnames=target.dimnames,
                                                                      source_size=source.size,
                                                                      source_dimnames=source.dimnames,
                                                                      target_index=target_index)

    def prepare(self):
        """
        Prepare the node for execution. It will create a SampleIdList of the
        correct size and prepare the outputs.
        """
        fastr.log.info('Preparing Node {} with size {} and dimnames {}'.format(self.id, self.outputsize, self.dimnames))

        if self.jobs is not None:
            raise exceptions.FastrNodeAreadyPreparedError('This Node has been previously prepared, cannot prepare '
                                                          'again as this will cause data loss!')

        self.jobs = {}

        for output in self.outputs.values():
            output.prepare()

    def _update(self, key, forward=True, backward=False):
        """
        Update the Node information and validity of the Node and propagate
         the update downstream. Updates inputs, inputgroups, outputsize and outputs.

        A Node is valid if:

        * All Inputs are valid (see :py:meth:`Input.update <fastr.core.inputoutput.Input.update>`)
        * All InputGroups are non-zero sized

        An Node is ready if:

        * The Node is valid
        * All Inputs are ready (see :py:meth:`Input.update <fastr.core.inputoutput.Input.update>`)

        """
        # Make sure the Inputs and input groups are up to date
        # fastr.log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.id))

        if backward:
            for sourced_node in self.get_sourced_nodes():
                sourced_node.update(key, False, backward)

        for input_ in self.inputs.values():
            input_.update(key, forward, backward)

        self.update_inputgroups()
        self._input_group_combiner.input_groups = self.inputgroups
        self._input_group_combiner.update()

        # Update own status
        valid = True
        ready = True
        messages = []

        for id_, input_ in self.inputs.items():
            if not input_.valid:
                valid = False
                for message in input_.messages:
                    messages.append('[{}] Input {} is not valid: {}'.format(self.id, input_.id, message))
            if not input_.ready:
                ready = False

        for input_group in self.inputgroups.values():
            if input_group.empty:
                valid = False
                messages.append('[{}] InputGroup {} is empty'.format(self.id, input_group.id))

        for id_, output in self.outputs.items():
            if output.resulting_datatype is not None:
                if not issubclass(output.resulting_datatype, DataType):
                    valid = False
                    messages.append('[{}] Output {} cannot determine the Output DataType (got {}), please specify a '
                                    'valid DataType or add casts to the Links'.format(self.id, id_, output.resulting_datatype))

        self._status['valid'] = valid
        self._status['messages'] = messages
        self._status['ready'] = (valid and ready)

        # Update all outputs
        for output in self.outputs.values():
            output.update(key, forward, backward)

        # Update all downstream listeners
        if forward:
            for listener in self.listeners:
                listener.update(key, forward, False)

    def update_inputgroups(self):
        """
        Update all input groups in this node
        """
        input_groups = OrderedDict()

        for input_ in self.inputs.values():
            if input_.input_group not in input_groups:
                input_groups[input_.input_group] = InputGroup(self, input_.input_group)
            input_groups[input_.input_group][input_.id] = input_

        self._input_groups = input_groups


class MacroNode(Node):
    """
    MacroNode encapsulates an entire network in a single node.
    """
    def __init__(self, network, id_=None, parent=None, cores=None, memory=None, walltime=None):
        """
        :param network: network to create macronode for
        :type network: Network
        """
        super(MacroNode, self).__init__(None, id_, parent=parent, cores=cores, memory=memory, walltime=walltime)

        # If macronode is loaded as a tool retrieve macro definition file(.py .xml .pickle .json) location
        if isinstance(network, Tool):
            # Check if Macro Definition in Tool is absolute Path or Relative
            if os.path.isabs(network.target.binary):
                network_path = network.target.binary
            else:
                network_location = os.path.dirname(network.filename)
                network_file = network.target.binary
                if network.target._paths[0]['type'] == 'bin':
                    network_relative_path = network.target._paths[0]['value']
                else:
                    network_relative_path = ''
                network_path = os.path.join(network_location, network_relative_path, network_file)
            # Check if macro definition exists.
            if not os.path.isfile(network_path):
                message = 'MacroNode \'{}\' definition file {} does not exist'.format(network.id, network_path)
                fastr.log.critical(message)
                raise exceptions.FastrTypeError(message)
            network = network_path

        # If network is an existing network
        if isinstance(network, fastr.Network):
            self._network = network
        # else if network is string(assume it is location to macro definition
        elif isinstance(network, (str, unicode)):
            # Check if file exists
            if not os.path.isfile(network):
                message = 'MacroNode definition file {} does not exist'.format(network)
                fastr.log.critical(message)
                raise exceptions.FastrTypeError(message)
            # If macro is python file
            if network.endswith('.py'):
                # py
                network_loader = imp.load_source('macro_node.utils', network)
                self._network = network_loader.main()
            # Else assume xml json pickkle
            else:
                # xml pickle, json, etc
                self._network = fastr.Network.loadf(network)
        # Else produce error
        else:
            message = 'Macro node should either be a Network a MacroTool or a FileName'
            fastr.log.critical(message)
            raise exceptions.FastrTypeError(message)

        try:
            self._network.is_valid()
            self._add_to_parent()
        except:
            message='Macro Node: {} is not a valid network'.format(id_)
            fastr.log.critical(message)
            raise exceptions.FastrValueError(message)

    def _add_to_parent(self):
        parent = self.parent
        id_ = self.id

        parent.add_stepid(id_, self)
        # Go through nodes and add them to the parent network
        for node in self._network.toolnodelist.itervalues():
            node._id = node.id + "__" + id_
            parent.add_stepid(id_, node)
            parent.add_node(node)

        # Go through source to set as macro node input
        for source_id, source in self._network.sourcelist.iteritems():
            self.inputs[source_id] = source.output.listeners[0].target.parent

        # Go through constants to set as non-required inputs
        for id, constant in self._network.constantlist.iteritems():
            self.inputs[id] = constant.output.listeners[0].target.parent
            constant._id = constant.id + "__" + id_
            parent.add_stepid(id_, constant)
            parent.add_node(constant)

        # Go through sinks to set as output
        for id, sink in self._network.sinklist.iteritems():
            self.outputs[id] = sink.input.get_sourced_outputs()[0]

        # Go through all links
        for id, link in self._network.linklist.iteritems():
            #if (not isinstance(link.source.node, SourceNode) and not isinstance(link.target.node,SinkNode)):
            source_node = link.source.node
            target_node = link.target.node
            if (not type(source_node) == SourceNode and not isinstance(target_node, SinkNode)):
                link.id = link.id + "__" + id_
                fastr.log.info('New link_id: {}'.format(link.id))
                link._parent = parent
                parent.add_link(link)

    def __getstate__(self):
        """
        Retrieve the state of the MacroNode

        :return: the state of the object
        :rtype dict:
        """
        state = super(MacroNode, self).__getstate__()
        state['network'] = self._network.__getstate__()
        return state

    def __setstate__(self, state):
        super(MacroNode, self).__setstate__(state)
        self._add_to_parent()

    def execute(self):
        raise exceptions.FastrNotImplementedError


class FlowNode(Node):
    """
    A Flow Node is a special subclass of Nodes in which the amount of samples
    can vary per Output. This allows non-default data flows.
    """
    _OutputType = Output

    def __init__(self, tool, id_=None, parent=None, cores=None, memory=None, walltime=None):
        """
        Instantiate a flow node.

        :param tool: The tool to base the node on
        :type tool: :py:class:`Tool <fastr.core.tool.Tool>`
        :param str id_: the id of the node
        :param parent: the parent network of the node
        :type parent: :py:class:`Network <fastr.core.network.Network>`
        :return: the newly created FlowNode
        """
        super(FlowNode, self).__init__(tool, id_, parent=parent, cores=cores, memory=memory, walltime=walltime)

        self._input_groups = OrderedDict()
        self.jobs = None

        # Update Inputs and self (which calls Outputs)
        self.update()

    @property
    def blocking(self):
        """
        A FlowNode is (for the moment) always considered blocking.

        :return: True
        """
        return True

    @property
    def outputsize(self):
        """
        Size of the outputs in this Node
        """
        # Get sizes of all input groups
        output_size = []
        for input_group in self.inputgroups.values():
            if input_group.size is not None:
                output_size.extend(input_group.size)
            else:
                return None

        output_size.append(sympy.symbols('N_{}'.format(self.id)))
        return tuple(output_size)

    @property
    def dimnames(self):
        """
        Names of the dimensions in the Node output. These will be reflected
        in the SampleIdList of this Node.
        """
        if self.nodegroup is not None:
            extra_dim = self.nodegroup
        else:
            extra_dim = self.id

        return super(FlowNode, self).dimnames + (extra_dim,)

    def set_result(self, job):
        """
        Incorporate result of a job into the FlowNode.

        :param Type job: job of which the result to store
        """
        fastr.log.debug('Job output data: {}'.format(job.output_data))

        # Get the main sample index from the Job
        sample_index = job.sample_index

        for output in self.outputs.values():
            if output.id not in job.output_data:
                fastr.log.error('Could not find expected data for {} in {}!'.format(output.fullid, job.output_data))
                continue

            output_data = job.output_data[output.id]

            fastr.log.debug('output_data = {}'.format(output_data))

            # Make sure dictionary is sorted, can also be list of items
            # which will be kept ordered
            if isinstance(output_data, dict):
                data = sorted(output_data.items())

            if not all(isinstance(x, (list, tuple)) and len(x) == 2 for x in data):
                raise exceptions.FastrValueError('The output data for a FlowNode should be a dictionary or a list of items (length 2 per entry)')

            for sample_nr, (sample_id, sample_data) in enumerate(data):
                orig_sample_id = sample_id

                # Ensure we have a SampleId (cast if need be)
                if not isinstance(sample_id, SampleId):
                    # Make sure sample_id is built from a tuple of str
                    if isinstance(sample_id, (str, unicode)):
                        sample_id = (str(sample_id),)
                    else:
                        sample_id = tuple(str(x) for x in sample_id)

                    sample_id = SampleId(sample_id)

                    fastr.log.debug('Change sample_id from {} ({}) to {} ({})'.format(orig_sample_id,
                                                                                      type(orig_sample_id).__name__,
                                                                                      sample_id,
                                                                                      type(sample_id).__name__))

                if len(sample_id) != output.ndims:
                    sample_id = job.sample_id + sample_id
                    fastr.log.debug('Updated sample_id to {}'.format(sample_id))
                    if len(sample_id) != output.ndims:
                        raise exceptions.FastrValueError('Sample ID {} has the wrong dimensionality!'.format(sample_id))

                fastr.log.debug('Setting data for blocking node: {} sample: {}'.format(output.fullid, sample_id))

                parsed_url = urlparse.urlparse(job.logurl)

                output_values = [(urlparse.urlunparse(('val',
                                                      parsed_url.netloc,
                                                      parsed_url.path,
                                                      parsed_url.params,
                                                      'outputname={}&nr={}&sampleid={}'.format(output.id,
                                                                                               c,
                                                                                               orig_sample_id),
                                                      ''))) for c, sample_data_value in enumerate(sample_data)]

                output_values = tuple(Deferred(x) for x in output_values)

                fastr.log.debug('Setting collected for {} sample_id {} sample_index {!r} data: {}'.format(output.fullid,
                                                                                                          sample_id,
                                                                                                          sample_index + (sample_nr),
                                                                                                          output_values))

                # Save with sample_index and sample nr in the extra dimension
                output[sample_id, sample_index + (sample_nr)] = SampleItem(sample_index + (sample_nr),
                                                                           sample_id,
                                                                           OrderedDict({0: tuple(output_values)}),
                                                                           {job})

                # Register the samples parent job
                self.jobs[sample_id] = job


class AdvancedFlowNode(FlowNode):
    _OutputType = AdvancedFlowOutput
    _JobType = InlineJob

    def execute(self):
        """
        Execute the node and create the jobs that need to run

        :return: list of jobs to run
        :rtype: list of :py:class:`Jobs <fastr.execution.job.Job>`
        """
        self.update(False, False)

        # Make sure a Node is valid
        if not self.valid:
            message = 'Node {} is not valid'.format(self.fullid)
            fastr.log.error(message)
            fastr.log.error('Messages:\n{}'.format('\n'.join(self.messages)))
            raise exceptions.FastrNodeNotValidError(message)
        inputgroups = self.inputgroups

        # Define output size and dimension names
        ig_masters = [ig.primary.id for ig in inputgroups.values()]
        fastr.log.debug('size: {} dimnames: {} masters: {}'.format(self.outputsize, self.dimnames, ig_masters))

        # Prepare the output of the Node
        fastr.log.debug('Preparing {} with size {} and dimnames {}'.format(self.fullid, self.outputsize, self.dimnames))
        self.prepare()

        fastr.log.debug('InputGroups: {}'.format(inputgroups.values()))
        fastr.log.debug('Inputs: {}'.format([x for ig in inputgroups.values() for x in ig.values()]))
        fastr.log.debug('Sources: {}'.format([x.source for ig in inputgroups.values() for x in ig.values()]))

        data = {x.id: x.items() for x in self.inputs.values()}
        target = self.tool.target

        job = self.create_job(SampleId('FLOW'),
                              SampleIndex(0),
                              job_data=data,
                              job_dependencies=None)

        with target:
            result = self.tool.interface.execute(target, data)

        job.flow_data = result.result_data

        output_data = {key: {str(v.id): v.data.sequence_part() for k, v in value.items()} for key, value in result.result_data.items()}
        job.output_data = output_data

        job.status = JobState.execution_done
        job.write()

        return [job]

    def set_result(self, job):
        for output, data in job.flow_data.items():
            fastr.log.debug('Advanced flow for output: {}'.format(output))
            for (sample_index, sample_id), value in data.items():
                fastr.log.debug('Advanced flow sample {!r} -> {}'.format(sample_index, list(value.data)))

                parsed_url = urlparse.urlparse(job.logurl)

                output_values = [(urlparse.urlunparse(('val',
                                                       parsed_url.netloc,
                                                       parsed_url.path,
                                                       parsed_url.params,
                                                       'outputname={}&nr={}&sampleid={}'.format(output,
                                                                                                c,
                                                                                                sample_id),
                                                       ''))) for c, sample_data_value in enumerate(value.data)]

                output_values = tuple(Deferred(x) for x in output_values)

                fastr.log.debug('Setting collected for {} sample_id {!r} sample_index {!r} data: {}'.format(output,
                                                                                                           sample_id,
                                                                                                           sample_index,
                                                                                                           output_values))

                # Save with sample_index and sample nr in the extra dimension
                self.outputs[output][sample_index] = SampleItem(value.index,
                                                                value.id,
                                                                OrderedDict({0: tuple(output_values)}),
                                                                {job})

        self.jobs['FLOW'] = job


class SourceNode(FlowNode):
    """
    Class providing a connection to data resources. This can be any kind of
    file, stream, database, etc from which data can be received.
    """

    __dataschemafile__ = 'SourceNode.schema.json'
    _OutputType = SourceOutput
    _JobType = SourceJob

    def __init__(self, datatype, id_=None):
        """
        Instantiation of the SourceNode.

        :param datatype: The (id of) the datatype of the output.
        :param id_: The url pattern.

        This class should never be instantiated directly (unless you know what
        you are doing). Instead create a source using the network class like
        shown in the usage example below.

        usage example:

        .. code-block:: python

          >>> import fastr
          >>> network = fastr.Network()
          >>> source = network.create_source(datatype=fastr.typelist['ITKImageFile'], id_='sourceN')
        """
        tool = fastr.toollist['Source']

        super(SourceNode, self).__init__(tool, id_)

        self._input_groups = []

        self.jobs = None

        # Set the DataType
        if datatype in fastr.typelist:
            if isinstance(datatype, str):
                datatype = fastr.typelist[datatype]
        else:
            message = 'Unknown DataType for SourceNode {} (found {}, which is not found in the typelist)!'.format(self.fullid, datatype)
            fastr.log.critical(message)
            raise exceptions.FastrValueError(message)

        self.datatype = datatype
        self._input_data = None
        self._outputsize = None
        self.outputsize = 'N_{}'.format(self.id)

    def __eq__(self, other):
        """Compare two Node instances with each other. This function ignores
        the parent and update status, but tests rest of the dict for equality.
        equality

        :param other: the other instances to compare to
        :type other: Node
        :returns: True if equal, False otherwise
        """
        if not isinstance(other, SourceNode):
            return NotImplemented

        dict_self = {k: v for k, v in self.__dict__.items()}
        del dict_self['_parent']
        del dict_self['_status']
        del dict_self['_input_groups']
        del dict_self['_input_data']
        del dict_self['_outputsize']
        del dict_self['_input_group_combiner']

        dict_other = {k: v for k, v in other.__dict__.items()}
        del dict_other['_parent']
        del dict_other['_status']
        del dict_other['_input_groups']
        del dict_other['_input_data']
        del dict_other['_outputsize']
        del dict_other['_input_group_combiner']

        return dict_self == dict_other

    def __getstate__(self):
        """
        Retrieve the state of the SourceNode

        :return: the state of the object
        :rtype dict:
        """
        state = super(SourceNode, self).__getstate__()

        return state

    def __setstate__(self, state):
        """
        Set the state of the SourceNode by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(SourceNode, self).__setstate__(state)

        self._input_data = None
        self._outputsize = None
        self.outputsize = 'N_{}'.format(self.id)

    @property
    def datatype(self):
        """
        The datatype of the data this source supplies.
        """
        return self.outputs['output'].datatype

    @datatype.setter
    def datatype(self, value):
        """
        The datatype of the data this source supplies. (setter)
        """
        self.outputs['output'].datatype = value

    @property
    def sourcegroup(self):
        fastr.log.warning('[DEPRECATED] The sourcegroup property of the'
                          ' SourceNode is deprecated and replaced by the'
                          ' nodegroup property of the Node. Please use that'
                          ' property instead, it will have the same'
                          ' functionality')
        return self.nodegroup

    @property
    def dimnames(self):
        """
        Names of the dimensions in the SourceNode output. These will be reflected
        in the SampleIdLists.
        """
        if self.nodegroup is not None:
            return self.nodegroup,
        else:
            return self.id,

    @property
    def output(self):
        """
        Shorthand for ``self.outputs['output']``
        """
        return self.outputs['output']

    @property
    def outputsize(self):
        """
        The size of output of this SourceNode
        """
        return self._outputsize

    @outputsize.setter
    def outputsize(self, value):
        # it seems pylint does not realize this is part of a property
        # pylint: disable=arguments-differ
        if isinstance(value, str):
            self._outputsize = (sympy.symbols(value),)
        elif isinstance(value, int):
            self._outputsize = (value,)
        else:
            try:
                self._outputsize = [x if isinstance(x, int) else sympy.symbols(x.replace(' ', '_')) for x in value]
            except TypeError:
                raise exceptions.FastrTypeError('Not a valid input type')

    @property
    def valid(self):
        """
        This does nothing. It only overloads the valid method of Node().
        The original is intended to check if the inputs are connected to
        some output. Since this class does not implement inputs, it is skipped.
        """
        return True

    def execute(self):
        """
        Execute the source node and create the jobs that need to run

        :return: list of jobs to run
        :rtype: list of :py:class:`Jobs <fastr.execution.job.Job>`
        """
        self.update(False, False)

        if not self.ready or self._input_data is None:
            msg = 'Cannot executed a SourceNode that is not ready! Messages:\n{}'.format('\n'.join(self.messages))
            fastr.log.error(msg)
            raise exceptions.FastrValueError(msg)

        joblist = []

        self.prepare()
        for index, (sampleid, value) in enumerate(self._input_data.items()):
            sample_index = SampleIndex(index)

            if all(not url.isurl(x) for x in value):
                # A simple string should not be send to IOPlugin for procesing
                fastr.log.debug('No job needed for sample {} at {}'.format(sampleid, self.fullid))
                self.jobs[sampleid] = None
                output_value = []

                for subvalue in value:
                    # it appears pylint does not realize that self.datatype is a class
                    # pylint: disable=not-callable
                    if self.datatype.isinstance(subvalue):
                        output_value.append(subvalue)
                    else:
                        output_value.append(self.datatype(subvalue))

                self.outputs['output'][sampleid, sample_index + (0,)] = SampleItem(sample_index + (0,),
                                                                                   sampleid,
                                                                                   {0: tuple(output_value)},
                                                                                   set())
            else:
                # We found an URL, should be
                fastr.log.debug('Spawning job for sample {} at {}'.format(sampleid, self.fullid))
                joblist.append(self.create_job(sampleid, sample_index, {'input': value}, []))

        return joblist

    def create_job(self, sample_id, sample_index, job_data, job_dependencies):
        job = super(SourceNode, self).create_job(sample_id, sample_index, job_data, job_dependencies)
        job._datatype = self.datatype.id
        return job

    def _wrap_arguments(self, job_data, sample_id, sample_index):
        """
        Wrap arguments into a list of tuples that the execution script can parse

        :param dict job_data: dictionary containing all input data for the job
        :param sample_id: the id of the corresponding sample
        :type sample_id: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the wrapped arguments in a tuple with the form ``(inputs, outputs)``

        .. note::
            For a SourceNode this function adds a few default (hidden) arguments
        """
        fastr.log.debug('Wrapping SourceNode with {}'.format(job_data))
        arguments = super(SourceNode, self)._wrap_arguments(job_data, sample_id, sample_index)
        arguments[0]['input'] = job_data['input']
        arguments[0]['behaviour'] = fastr.typelist['__source-interface__behaviour__Enum__']('source'),
        arguments[0]['datatype'] = fastr.typelist['String'](self.datatype.id),
        arguments[0]['sample_id'] = fastr.typelist['String'](str(sample_id)),

        outputurl = '{}/{}/{}/result'.format(self.parent.tmpurl, self.id, '__'.join(sample_id))
        outputpath = fastr.vfs.url_to_path(outputurl)
        if not os.path.exists(outputpath):
            os.makedirs(outputpath)
        arguments[0]['targetdir'] = fastr.typelist['Directory'](outputurl),

        return arguments

    def set_data(self, data, ids=None):
        """
        Set the data of this source node.

        :param data: the data to use
        :type data: dict, OrderedDict or list of urls
        :param ids: if data is a list, a list of accompanying ids
        """
        self._input_data = OrderedDict()

        # Check if data has key or generate keys
        fastr.log.debug('Storing {} (ids {}) in {}'.format(data, ids, self.fullid))
        if isinstance(data, dict):
            # Have data sorted on ids
            ids, data = zip(*sorted(data.items()))
            ids = [SampleId(x) for x in ids]
        elif isinstance(data, OrderedDict):
            ids, data = data.keys(), data.values()
        elif isinstance(data, list):
            if ids is None:
                ids = [SampleId('id_{}'.format(k)) for k in range(len(data))]
            elif not isinstance(ids, list):
                raise exceptions.FastrTypeError('Invalid type! The ids argument should be a list that matches the data samples!')
        elif isinstance(data, tuple):
            # A single sample with cardinality
            ids = [SampleId('id_0')]
            data = [data]
        else:
            ids = [SampleId('id_0')]
            data = [data]

        fastr.log.debug('Set data in {} with {} (Type {})'.format(self.id, data, self.datatype))

        for key, value in zip(ids, data):
            if isinstance(value, tuple):
                self._input_data[key] = tuple(x if self.datatype.isinstance(x) else str(x) for x in value)
            else:
                self._input_data[key] = (value if self.datatype.isinstance(value) else str(value)),
            fastr.log.debug('Result {}: {} (Type {})'.format(key, self._input_data[key], type(self._input_data[key]).__name__))

        self._status['ready'] = True
        self.outputsize = len(self._input_data),

    def _update(self, key, forward=True, backward=False):
        """
        Update the Node information and validity of the Node and propagate
         the update downstream. Updates inputs, inputgroups, outputsize and outputs.

        A Node is valid if:

        * All Inputs are valid (see :py:meth:`Input.update <fastr.core.inputoutput.Input.update>`)
        * All InputGroups are non-zero sized

        An Node is ready if:

        * The Node is valid
        * All Inputs are ready (see :py:meth:`Input.update <fastr.core.inputoutput.Input.update>`)

        """
        # Make sure the Inputs and input groups are up to date
        # fastr.log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.id))

        for input_ in self.inputs.values():
            input_.update(key)

        self.update_inputgroups()

        # Update own status
        valid = True
        ready = True
        messages = []

        for id_, input_ in self.inputs.items():
            if not input_.valid:
                valid = False
                for message in input_.messages:
                    messages.append('Input {} is not valid: {}'.format(id_, message))
            if not input_.ready:
                ready = False

        for input_group in self.inputgroups.values():
            if input_group.empty:
                valid = False
                messages.append('InputGroup {} is empty'.format(input_group.id))

        for id_, output in self.outputs.items():
            if output.resulting_datatype is not None:
                if not issubclass(output.resulting_datatype, DataType):
                    valid = False
                    messages.append(
                        'Output {} cannot determine the Output DataType (got {}), '
                        'please specify a valid DataType or add casts to the Links'.format(
                            id_, output.resulting_datatype))

        self._status['valid'] = valid
        self._status['messages'] = messages
        self._status['ready'] = (valid and ready)

        # Update all outputs
        for output in self.outputs.values():
            output.update(key)

        # Update all downstream listeners
        if forward:
            for listener in self.listeners:
                listener.update(key, forward, backward)


class SinkNode(Node):
    """
    Class which handles where the output goes. This can be any kind of file, e.g.
    image files, textfiles, config files, etc.
    """

    __dataschemafile__ = 'SinkNode.schema.json'
    _JobType = SinkJob

    def __init__(self, datatype, id_=None):
        """ Instantiation of the SourceNode.

        :param datatype: The datatype of the output.
        :param id_: the id of the node to create
        :return: newly created sink node

        usage example:

        .. code-block:: python

          >>> import fastr
          >>> network = fastr.Network()
          >>> sink = network.create_sink(datatype=fastr.typelist['ITKImageFile'], id_='SinkN')

        """
        Node.__init__(self, fastr.toollist['Sink'], id_)
        # Set the DataType
        if datatype in fastr.typelist:
            if isinstance(datatype, str):
                datatype = fastr.typelist[datatype]
        else:
            message = 'Invalid DataType for SinkNode {} (found {})!'.format(self.fullid, datatype)
            fastr.log.critical(message)
            raise exceptions.FastrValueError(message)

        self.datatype = datatype
        # TODO: this code cannot function, need to find a work-around
        #self._tool.inputs['input'].datatype = datatype
        self.url = None

    def __getstate__(self):
        state = super(SinkNode, self).__getstate__()
        state['url'] = self.url
        return state

    def __setstate__(self, state):
        super(SinkNode, self).__setstate__(state)
        self.url = state['url']

    @property
    def datatype(self):
        """
        The datatype of the data this sink can store.
        """
        return self.inputs['input'].datatype

    @datatype.setter
    def datatype(self, value):
        """
        The datatype of the data this sink can store (setter).
        """
        self.inputs['input'].datatype = value

    @property
    def input(self):
        """
        The default input of the sink Node
        """
        return self.inputs['input']

    @input.setter
    def input(self, value):
        """
        The default input of the sink Node (setter)
        """
        self.inputs['input'] = value

    def execute(self):
        """
        Execute the sink node and create the jobs that need to run

        :return: list of jobs to run
        :rtype: list of :py:class:`Jobs <fastr.execution.job.Job>`
        """
        self.update(False, False)

        joblist = []

        self.prepare()

        for sample_index, sampleid, data, jobs in self.inputs['input'].iteritems():
            for cardinality_nr, value in enumerate(data.sequence_part()):
                fastr.log.debug('Spawning job for {}'.format(self.inputs['input'].fullid))
                joblist.append(self.create_job(sampleid, sample_index, {'input': SampleItem(sample_index, sampleid, SampleValue({0: (value,)})), 'cardinality': cardinality_nr}, jobs))

        return joblist

    def set_data(self, data):
        """
        Set the targets of this sink node.

        :param data: the targets rules for where to write the data
        :type data: dict or list of urls

        The target rules can include a few fields that can be filled out:

        =========== ==================================================================
        field       description
        =========== ==================================================================
        sample_id   the sample id of the sample written in string form
        cardinality the cardinality of the sample written
        ext         the extension of the datatype of the written data, including the .
        network     the id of the network the sink is part of
        node        the id of the node of the sink
        timestamp   the iso formatted datetime the network execution started
        uuid        the uuid of the network run (generated using uuid.uuid1)
        =========== ==================================================================

        An example of a valid target could be:

        .. code-block:: python

          >>> target = 'vfs://output_mnt/some/path/image_{sample_id}_{cardinality}{ext}'

        """
        if isinstance(data, (str, unicode)):
            try:
                data.format(sample_id='dummy',
                            cardinality=0,
                            ext='.ext',
                            network='network',
                            node='node',
                            timestamp='timestamp',
                            uuid='uuid')
            except KeyError as error:
                raise exceptions.FastrValueError('Using unknown substitution "{}" in SinkData "{}", valid substitution fields are: sample_id, cardinality, ext'.format(error.message, data))
            self.url = data
            self._status['ready'] = True
        else:
            raise exceptions.FastrTypeError('Invalid datatype for SinkNode data, expected str but got {}!'.format(type(data).__name__))

    def create_job(self, sample_id, sample_index, job_data, job_dependencies):
        """
        Create a job for a sink based on the sample id, job data and job dependencies.

        :param sample_id: the id of the corresponding sample
        :type sample_id: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :param dict job_data: dictionary containing all input data for the job
        :param job_dependencies: other jobs that need to finish before this job can run
        :return: the created job
        :rtype: :py:class:`Job <fastr.execution.job.Job>`
        """

        # Create new jobid
        jobid = '{}___{}___{}___{}'.format(self.parent.id, self.id, sample_id, job_data['cardinality'])

        # Determine new tmp dir
        outputurl = '{}/{}/{}__{}'.format(self.parent.tmpurl, self.id, sample_id, job_data['cardinality'])

        substitutions = {'sample_id': sample_id,
                         'cardinality': job_data['cardinality'],
                         'timestamp': self.parent.timestamp.isoformat(),
                         'uuid': self.parent.uuid,
                         'network': self.parent.id,
                         'node': self.id}

        job = super(SinkNode, self).create_job(sample_id, sample_index, job_data, job_dependencies, jobid=jobid, outputurl=outputurl, substitutions=substitutions)

        self.jobs[sample_id] = job
        return job

    def _wrap_arguments(self, job_data, sample_id, sample_index):
        """
        Wrap arguments into a list of tuples that the execution script can parse

        :param dict job_data: dictionary containing all input data for the job
        :param sample_id: the id of the corresponding sample
        :type sample_id: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the wrapped arguments in a tuple with the form ``(inputs, outputs)``

        .. note::
            For a SinkNode this function adds a few default (hidden) arguments
        """
        arguments = super(SinkNode, self)._wrap_arguments(job_data, sample_id, sample_index)
        arguments[0]['behaviour'] = fastr.typelist['__source-interface__behaviour__Enum__']('sink'),
        arguments[0]['output'] = fastr.typelist['String'](self.url),
        arguments[0]['datatype'] = fastr.typelist['String'](self.datatype.id),

        fastr.log.debug('Wrapped Sink arguments to {}'.format(arguments))
        return arguments


class ConstantNode(SourceNode):
    """
    Class encapsulating one output for which a value can be set. For example
    used to set a scalar value to the input of a node.
    """

    __dataschemafile__ = 'ConstantNode.schema.json'

    def __init__(self, datatype, data, id_=None):
        """
        Instantiation of the ConstantNode.

        :param datatype: The datatype of the output.
        :param data: the prefilled data to use.
        :param id_: The url pattern.

        This class should never be instantiated directly (unless you know what
        you are doing). Instead create a constant using the network class like
        shown in the usage example below.

        usage example:

        .. code-block:: python

          >>> import fastr
          >>> network = fastr.Network()
          >>> source = network.create_source(datatype=fastr.typelist['ITKImageFile'], id_='sourceN')

        or alternatively create a constant node by assigning data to an item in an InputDict:

        .. code-block:: python

          >>> node_a.inputs['in'] = ['some', 'data']

        which automatically creates and links a ConstantNode to the specified Input
        """
        super(ConstantNode, self).__init__(datatype, id_)
        self.set_data(data)
        self._data = self._input_data

    def __getstate__(self):
        """
        Retrieve the state of the ConstantNode

        :return: the state of the object
        :rtype dict:
        """
        state = super(ConstantNode, self).__getstate__()

        state['data'] = self._data.items()

        return state

    def __setstate__(self, state):
        """
        Set the state of the ConstantNode by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(ConstantNode, self).__setstate__(state)

        self._data = OrderedDict((SampleId(str(x) for x in key), tuple(str(x) for x in value)) for key, value in state['data'])
        self.set_data()  # Make sure that the output size etc gets set

    def set_data(self, data=None, ids=None):
        """
        Set the data of this constant node in the correct way. This is mainly
        for compatibility with the parent class SourceNode

        :param data: the data to use
        :type data: dict or list of urls
        :param ids: if data is a list, a list of accompanying ids
        """
        # We have to arguments to match the superclas
        # pylint: disable=unused-argument
        if data is None and self.data is not None:
            self._input_data = self.data
        else:
            super(ConstantNode, self).set_data(data, ids)

    @property
    def data(self):
        """
        The data stored in this constant node
        """
        return self._data

    def execute(self):
        """
        Execute the constant node and create the jobs that need to run

        :return: list of jobs to run
        :rtype: list of :py:class:`Jobs <fastr.execution.job.Job>`
        """
        # Make sure the data is set
        self.set_data()

        # Run as a normal SourceNode
        return super(ConstantNode, self).execute()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the XNATStorage plugin for fastr
"""

import fnmatch
from itertools import izip
import netrc
import os
import re
import tempfile
import urllib
import urlparse
from zipfile import ZipFile
from collections import OrderedDict

import xnat

import fastr
from fastr import exceptions
from fastr.core.ioplugin import IOPlugin


class XNATStorage(IOPlugin):

    """
    .. warning::

        As this IOPlugin is under development, it has not been thoroughly
        tested.

    The XNATStorage plugin is an IOPlugin that can download data from and
    upload data to an XNAT server. It uses its own ``xnat://`` URL scheme.
    This is a scheme specific for this plugin and though it looks somewhat
    like the XNAT rest interface, a different type or URL.

    Data resources can be access directly by a data url::

        xnat://xnat.example.com/data/archive/projects/sandbox/subjects/subject001/experiments/experiment001/scans/T1/resources/DICOM
        xnat://xnat.example.com/data/archive/projects/sandbox/subjects/subject001/experiments/*_BRAIN/scans/T1/resources/DICOM

    In the second URL you can see a wildcard being used. This is possible at
    long as it resolves to exactly one item.

    The ``id`` query element will change the field from the default experiment to
    subject and the ``label`` query element sets the use of the label as the
    fastr id (instead of the XNAT id) to ``True`` (the default is ``False``)

    To disable ``https`` transport and use ``http`` instead the query string
    can be modified to add ``insecure=true``. This will make the plugin send
    requests over ``http``::

        xnat://xnat.example.com/data/archive/projects/sandbox/subjects/subject001/experiments/*_BRAIN/scans/T1/resources/DICOM?insecure=true

    For sinks it is import to know where to save the data. Sometimes you want
    to save data in a new assessor/resource and it needs to be created. To
    allow the Fastr sink to create an object in XNAT, you have to supply the
    type as a query parameter::

        xnat://xnat.bmia.nl/data/archive/projects/sandbox/subjects/S01/experiments/_BRAIN/assessors/test_assessor/resources/IMAGE/files/image.nii.gz?resource_type=xnat:resourceCatalog&assessor_type=xnat:qcAssessmentData

    Valid options are: subject_type, experiment_type, assessor_type, scan_type,
    and resource_type.

    If you want to do a search where
    multiple resources are returned, it is possible to use a search url::

        xnat://xnat.example.com/search?projects=sandbox&subjects=subject[0-9][0-9][0-9]&experiments=*_BRAIN&scans=T1&resources=DICOM

    This will return all DICOMs for the T1 scans for experiments that end with _BRAIN that belong to a
    subjectXXX where XXX is a 3 digit number. By default the ID for the samples
    will be the experiment XNAT ID (e.g. XNAT_E00123). The wildcards that can
    be the used are the same UNIX shell-style wildcards as provided by the
    module :py:mod:`fnmatch`.

    It is possible to change the id to a different fields id or label. Valid
    fields are project, subject, experiment, scan, and resource::

        xnat://xnat.example.com/search?projects=sandbox&subjects=subject[0-9][0-9][0-9]&experiments=*_BRAIN&scans=T1&resources=DICOM&id=subject&label=true

    The following variables can be set in the search query:

    ============= ============== =============================================================================================
    variable      default        usage
    ============= ============== =============================================================================================
    projects      ``*``          The project(s) to select, can contain wildcards (see :py:mod:`fnmatch`)
    subjects      ``*``          The subject(s) to select, can contain wildcards (see :py:mod:`fnmatch`)
    experiments   ``*``          The experiment(s) to select, can contain wildcards (see :py:mod:`fnmatch`)
    scans         ``*``          The scan(s) to select, can contain wildcards (see :py:mod:`fnmatch`)
    resources     ``*``          The resource(s) to select, can contain wildcards (see :py:mod:`fnmatch`)
    id            ``experiment`` What field to use a the id, can be: project, subject, experiment, scan, or resource
    label         ``false``      Indicate the XNAT label should be used as fastr id, options ``true`` or ``false``
    insecure      ``false``      Change the url scheme to be used to http instead of https
    regex         ``false``      Change search to use regex :py:func:`re.match` instead of fnmatch for matching
    ============= ============== =============================================================================================

    For storing credentials the ``.netrc`` file can be used. This is a common
    way to store credentials on UNIX systems. It is required that the file is
    only accessible by the owner only or a ``NetrcParseError`` will be raised.
    A netrc file is really easy to create, as its entries look like::

        machine xnat.example.com
                login username
                password secret123

    See the :py:mod:`netrc module <netrc>` or the
    `GNU inet utils website <http://www.gnu.org/software/inetutils/manual/html_node/The-_002enetrc-file.html#The-_002enetrc-file>`_
    for more information about the ``.netrc`` file.

    .. note::

        On windows the location of the netrc file is assumed to be
        ``os.path.expanduser('~/_netrc')``. The leading underscore is
        because windows does not like filename starting with a dot.

    .. note::

        For scan the label will be the scan type (this is initially
        the same as the series description, but can be updated manually
        or the XNAT scan type cleanup).

    .. warning::

        labels in XNAT are not guaranteed to be unique, so be careful
        when using them as the sample ID.

    For background on XNAT, see the
    `XNAT API DIRECTORY <https://wiki.xnat.org/display/XNAT16/XNAT+REST+API+Directory>`_
    for the REST API of XNAT.
    """
    scheme = 'xnat'

    def __init__(self):
        # initialize the instance and register the scheme
        super(XNATStorage, self).__init__()
        self._xnat = (None, None)

    def cleanup(self):
        if self.xnat is not None:
            self.xnat.disconnect()

    @property
    def server(self):
        return self._xnat[0]

    @property
    def xnat(self):
        return self._xnat[1]

    def connect(self, server, path='', insecure=False):
        if self.server != server:
            # Try to neatly clean previous connection
            if self.xnat is not None:
                self.xnat.disconnect()

            try:
                netrc_file = os.path.join('~', '_netrc' if os.name == 'nt' else '.netrc')
                netrc_file = os.path.expanduser(netrc_file)
                user, _, password = netrc.netrc(netrc_file).authenticators(server)
            except TypeError:
                raise exceptions.FastrValueError('Could not retrieve login info for "{}" from the .netrc file!'.format(server))

            # Create the URL for the XNAT connection
            schema = 'http' if insecure else 'https'
            session = xnat.connect(urlparse.urlunparse([schema, server, path, '', '', '']), user=user, password=password, debug=True)

            self._xnat = (server, session)

    @staticmethod
    def _path_to_dict(path):
        fastr.log.info('Converting {} to dict...'.format(path))
        if not path.startswith('/data/archive/'):
            # prefix, path = path.split('/data/archive')
            raise ValueError('Resources to be located should have a path starting with /data/ (found {})'.format(path))

        # Break path apart
        parts = path.lstrip('/').split('/', 13)

        # Ignore first two parts and build a dict from /key/value/key/value pattern
        path_iterator = parts[2:].__iter__()
        location = OrderedDict()
        for key, value in izip(path_iterator, path_iterator):
            if key == 'files':
                filepath = [value] + list(path_iterator)
                value = '/'.join(filepath)

            location[key] = value

        fastr.log.info('Found {}'.format(location))
        return location

    def _locate_resource(self, url, create=False, use_regex=False):
        resources = self._find_objects(url=url, create=create, use_regex=use_regex)

        if len(resources) == 0:
            raise ValueError('Could not find data object at {}'.format(url))
        elif len(resources) > 1:
            raise ValueError('Data item does not point to a unique resource! Matches found: {}'.format([x.fulluri for x in resources]))

        resource = resources[0]

        # Make sure the return value is actually a resource
        resource_cls = self.xnat.XNAT_CLASS_LOOKUP['xnat:abstractResource']
        if not isinstance(resource, resource_cls):
            raise TypeError('The resource should be an instance of {}'.format(resource_cls))

        return resource

    def fetch_url(self, inurl, outpath):
        """
        Get the file(s) or values from XNAT.

        :param inurl: url to the item in the data store
        :param outpath: path where to store the fetch data locally
        :param datatype: the DataType of the retrieved URL
        """

        if fastr.data.url.get_url_scheme(inurl) != self.scheme:
            raise exceptions.FastrValueError('URL not of {} type!'.format(self.scheme))

        # Create a session for this retrieval
        url = urlparse.urlparse(inurl)
        path_prefix = url.path[:url.path.find('/data/archive')]
        url = url._replace(path=url.path[len(path_prefix):])  # Strip the prefix of the url path
        inurl = urlparse.urlunparse(url)

        if not url.scheme == 'xnat':
            raise ValueError('URL does not has an xnat scheme')

        if not url.path.startswith('/data/archive'):
            raise ValueError('Can only fetch urls with the /data/archive path')

        insecure = urlparse.parse_qs(url.query).get('insecure', ['0'])[0] in ['true', '1']
        use_regex = urlparse.parse_qs(url.query).get('regex', ['0'])[0] in ['true', '1']
        self.connect(url.netloc, path=path_prefix, insecure=insecure)

        # Find the filepath within the resource
        location = self._path_to_dict(url.path)
        filepath = location.get('files', '')

        # Find the resource
        resource = self._locate_resource(inurl, use_regex=use_regex)

        # Download the Resource
        workdir = outpath
        if not os.path.isdir(workdir):
            os.makedirs(workdir)

        # Create uniquer dir to download in
        workdir = tempfile.mkdtemp(prefix='fastr_xnat_{}_tmp'.format(resource.id), dir=outpath)
        resource.download_dir(workdir)

        # Determine data path
        data_path = os.path.join(workdir,
                                 os.listdir(workdir)[0],  # This should be safe as there should be only one experiment
                                 'scans', '{}-{}'.format(resource.data['cat_id'], "".join([c if c.isalnum() else "_" for c in resource.data['cat_desc']])),
                                 'resources',
                                 resource.label.replace(' ', '_'),
                                 'files',
                                 filepath)
        fastr.log.debug('Data located in: {}'.format(data_path))
        return data_path

    def put_url(self, inpath, outurl):
        """
        Upload the files to the XNAT storage

        :param inpath: path to the local data
        :param outurl: url to where to store the data in the external data store.
        """
        # Create a session for this retrieval
        url = urlparse.urlparse(outurl)
        path_prefix = url.path[:url.path.find('/data/archive')]
        url = url._replace(path=url.path[len(path_prefix):])  # Strip the prefix of the url path
        outurl = urlparse.urlunparse(url)

        insecure = urlparse.parse_qs(url.query).get('insecure', ['false'])[0] in ['true', '1']
        use_regex = urlparse.parse_qs(url.query).get('regex', ['0'])[0] in ['true', '1']
        self.connect(url.netloc, path=path_prefix, insecure=insecure)

        # Determine the resource to upload to
        resource = self._locate_resource(outurl, create=True, use_regex=use_regex)

        # Determine the file within xnat
        parsed_url = urlparse.urlparse(outurl)
        location = self._path_to_dict(parsed_url.path)

        # Upload the file
        fastr.log.info('Uploading to: {}'.format(resource.fulluri))
        fastr.log.info('Uploading to path: {}'.format(location['files']))
        try:
            resource.upload(inpath, location['files'])
            return True
        except xnat.exceptions.XNATUploadError as exception:
            fastr.log.error('Encountered error when uploading data: {}'.format(exception))
            return False

    def _find_objects(self, url, create=False, use_regex=False):
        fastr.log.info('Locating {}'.format(url))
        # Parse url
        parsed_url = urlparse.urlparse(url)
        path = parsed_url.path
        query = urlparse.parse_qs(parsed_url.query)

        if not path.startswith('/data/archive'):
            raise ValueError('Resources to be located should have a path starting with /data/archive')

        # Create a search uri
        location = self._path_to_dict(path)

        if 'resources' not in location:
            raise ValueError('All files should be located inside a resource, did not'
                             ' find resources level in {}'.format(location))

        # Sort xsi type directives neatly
        types = {}
        for key in location.keys():
            option1 = (key.rstrip('s') + '_type')
            option2 = (key + '_type')
            if option1 in query:
                types[key] = query[option1][0]
            if option2 in query:
                types[key] = query[option2][0]

        items = None
        # Parse location part by part
        for object_type, object_key in location.items():
            # We don't want to go into files, those are put but not created
            # and they get via the resources
            if object_type == 'files':
                break

            fastr.log.info('Locating {} / {} in {}'.format(object_type, object_key, items))
            new_items = self._resolve_url_part(object_type, object_key, use_regex=use_regex, parents=items)

            if len(new_items) == 0:
                if not create:
                    raise ValueError('Could not find data parent_object at {} (No values at level {})'.format(url,
                                                                                                              object_type))
                elif items is not None and len(items) == 1:
                    fastr.log.debug('Items: {}'.format(items))
                    parent_object = items[0]

                    # Get the required xsitype
                    if object_type in types:
                        xsi_type = types[object_type]
                    else:
                        raise ValueError('Could not find the correct xsi:type for {} (available hints: {})'.format(object_type,
                                                                                                                   types))

                    if '*' in object_key or '?' in object_key or '[' in object_key or ']' in object_key:
                        raise ValueError('Illegal characters found in name of object_key'
                                         ' to create! (characters ?*[] or illegal!), found: {}'.format(object_key))

                    fastr.log.info('Creating new object under {} with type {}'.format(parent_object.uri, xsi_type))

                    # Create the object with the correct secondary lookup
                    cls = self.xnat.XNAT_CLASS_LOOKUP[xsi_type]
                    kwargs = {cls.SECONDARY_LOOKUP_FIELD: object_key}
                    cls(parent=parent_object, **kwargs)

                    new_items = self._resolve_url_part(object_type, object_key, use_regex=use_regex, parents=items)

                    if len(new_items) != 1:
                        raise ValueError('There appears to be a problem creating the object_key!')
                else:
                    raise ValueError('To create an object, the path should point to a unique parent'
                                     ' object! Found {} matching items'.format(len(items)))
            # Accept the new items for the new level scan
            items = new_items

        return items

    def _resolve_url_part(self, level, query=None, use_regex=False, parents=None):
        """
        Get all matching projects

        :param dict query: the query to find projects to match for
        :return:
        """
        # If there are no parents, work directly on the session
        if parents is None:
            parents = [self.xnat]

        if query is None:
            query = '.*' if use_regex else '*'

        fastr.log.info('Find {}: {} (parents: {})'.format(level, query, parents))

        # Get all objects
        objects = []
        for parent in parents:
            extra_objects = getattr(parent, level)
            if use_regex:
                fastr.log.info('EXTRA REGEX CANDIDATES: {} [query: {}]'.format(extra_objects.values(), query))
                extra_objects = [x for x in extra_objects.values() if re.match(query, getattr(x, extra_objects.secondary_lookup_field)) or x.id == query]
            else:
                fastr.log.info('EXTRA CANDIDATES: {}'.format(extra_objects.values()))
                extra_objects = [x for x in extra_objects.values() if fnmatch.fnmatchcase(getattr(x, extra_objects.secondary_lookup_field), query) or x.id == query]
            objects.extend(extra_objects)

        return objects

    def expand_url(self, url):
        # Check if there is a wildcard in the URL:
        parsed_url = urlparse.urlparse(url)

        if parsed_url.path == '/search':
            # Parse the query
            query = urlparse.parse_qs(parsed_url.query)

            # Check if all fields given are valid fieldnames
            valid_fields = ("projects",
                            "subjects",
                            "experiments",
                            "scans",
                            "resources",
                            "id",
                            "label",
                            "insecure",
                            "regex")

            valid_query = True
            for key in query.keys():
                if key not in valid_fields:
                    fastr.log.error('Using invalid query field {} options are {}!'.format(key,
                                                                                            valid_fields))
                    valid_query = False

            if not valid_query:
                raise ValueError('The query was malformed, see the error log for details.')

            id_field = query.get('id', ['experiment'])[0]
            id_use_label = query.get('label', ['0'])[0].lower() in ['1', 'true']
            use_regex = query.get('regex', ['0'])[0].lower() in ['1', 'true']
            insecure = query.get('insecure', ['0'])[0] in ['true', '1']

            # Make sure we are connect to the correct server
            self.connect(parsed_url.netloc, insecure=insecure)

            if id_field not in ['project', 'subject', 'experiment', 'scan', 'resource']:
                raise exceptions.FastrValueError('Requested id field ({}) is not a valid option!'.format(id_field))

            # Create the url version for the search
            default = ['.*'] if use_regex else ['*']
            search_path = '/data/archive/projects/{p}/subjects/{s}/experiments/{e}/scans/{sc}/resources/{r}'.format(
                p=query.get('projects', default)[0],
                s=query.get('subjects', default)[0],
                e=query.get('experiments', default)[0],
                sc=query.get('scans', default)[0],
                r=query.get('resources', default)[0]
            )

            # Find all matching resources
            resources = self._find_objects(search_path, use_regex=use_regex)

            # Format the new expanded urls
            urls = []
            for resource in resources:
                match = re.match('/data/experiments/(?P<experiment>[a-zA-Z0-9_]+)/scans/(?P<scan>[a-zA-Z0-9_]+)/resources/(?P<resource>[a-zA-Z0-9_]+)', resource.uri)
                experiment = self.xnat.experiments[match.group('experiment')]
                project = self.xnat.projects[experiment.project]
                subject = self.xnat.subjects[experiment.subject_id]
                scan = experiment.scans[match.group('scan')]

                # TODO: Just use resource.fulluri + '/files/filename' ?
                newpath = '/data/archive/projects/{}/subjects/{}/experiments/{}/scans/{}/resources/{}/files/{}'.format(
                    project.id,
                    subject.id,
                    experiment.id,
                    scan.id,
                    resource.id,
                    query.get('files', [''])[0]
                )

                newurl = urlparse.urlunparse(('xnat', parsed_url.netloc, newpath, parsed_url.params, '', ''))

                # Determine the ID of the sample
                if id_field == 'resource':
                    id_obj = resource
                elif id_field == 'scan':
                    id_obj = scan
                elif id_field == 'experiment':
                    id_obj = experiment
                elif id_field == 'subject':
                    id_obj = subject
                else:
                    # Must be project
                    id_obj = project

                if id_use_label:
                    id_ = id_obj.label
                else:
                    id_ = id_obj.id

                urls.append((id_, newurl))

            return tuple(urls)
        else:
            return url



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the core class for all managers
"""

import collections
import os
import re
import sys
import traceback
from abc import ABCMeta, abstractmethod, abstractproperty
from fastr.exceptions import FastrKeyError, FastrNotImplementedError
import fastr


class BaseManager(collections.MutableMapping):
    """
    Baseclass for a Manager, subclasses needs to override the following methods:
      BaseManager._item_extension, BaseManager._load_item()

    .. automethod:: _item_extension

    .. automethod:: _load_item
    """
    __metaclass__ = ABCMeta

    def __init__(self, path=None, recursive=False):
        """
        The BaseManager constructor

        :param path: path to scan for items, or None for no path
        :type path: str or None
        :param bool recursive: Flag to indicate a recursive search is desired
        :return: the newly created BaseManager
        :rtype: BaseManager
        """
        self._scanned_directories = []
        self._path = (path, recursive)
        self._data = None

    @property
    def data(self):
        """
        The actual data dict underlying this Manager
        """
        if self._data is None:
            self.populate()

        return self._data

    def populate(self):
        """
        Populate the manager with the data. This is a method that will be
        called when the Managers data is first accessed. This way we avoid
        doing expensive directory scans when the data is never requested.
        """
        self._data = dict()
        if self._path[0] is not None:
            self._scan_directory(self._path[0], recursive=self._path[1])

    def reload(self):
        """
        Reload entire contents of this manager.
        """
        self.clear()
        self.populate()

    def __repr__(self):
        """
        Convert the BaseManager to a representation string.

        :return: Representation string
        :rtype: str
        """
        if len(self) == 0:
            return 'Empty {}'.format(type(self).__name__)

        keylist = [self._print_key(x) for x in self.keys()]
        if isinstance(keylist[0], tuple):
            width = zip(*keylist)
            width = [max([len(str(x)) for x in y]) for y in width]
            format_base_key = '  '.join(['{{key[{n}]:<{width}}}'.format(n=n, width=w) for n, w in enumerate(width)])
        else:
            width = max([len(str(x)) for x in keylist if x is not None])
            format_base_key = '{{key:<{width}}}'.format(width=width)

        valuelist = [self._print_value(x) for x in self.values()]
        if isinstance(valuelist[0], tuple):
            width = zip(*valuelist)
            width = [max([len(str(x)) for x in y]) for y in width]
            format_base_value = '  '.join(['{{val[{n}]:<{width}}}'.format(n=n, width=w) for n, w in enumerate(width)])
        else:
            width = max([len(str(x)) for x in valuelist])
            format_base_value = '{{val:<{width}}}'.format(width=width)

        output = [type(self).__name__]
        format_base = '{}  :  {}'.format(format_base_key, format_base_value)

        for key, val in sorted(self.items()):
            key = self._print_key(key)
            if key is not None:
                output.append(format_base.format(key=key, val=self._print_value(val)))

        return '\n'.join(output)

    def __getitem__(self, key):
        """
        Retrieve item from BaseManager

        :param key: the key of the item to retrieve
        :return: the value indicated by the key
        :raises FastrKeyError: if the key is not found in the BaseManager
        """
        transformed_key = self.__keytransform__(key)
        if transformed_key not in self.data:
            raise FastrKeyError('Object "{}" ("{}") not found in {}'.format(key,
                                                                            transformed_key,
                                                                            type(self).__name__))

        return self.data[transformed_key]

    def __setitem__(self, key, value):
        """
        Set item in the BaseManager

        :param key: the key of the item to store
        :param value: the value of the item to store
        :return: None
        """
        self.data[self.__keytransform__(key)] = value

    def __delitem__(self, key):
        """
        Remove item from the BaseManager

        :param key: key of the item to remove
        :return: None
        :raises FastrKeyError: if the key is not found in the BaseManager
        """

        if key not in self.data:
            raise FastrKeyError('Plugin "{}" not found in {}'.format(key, type(self).__name__))

        del self.data[self.__keytransform__(key)]

    def __iter__(self):
        """
        Get an iterator from the BaseManager. The iterator will iterate over
        the keys of the BaseManager.

        :return: the iterator
        :rtype: dictionary-keyiterator
        """
        return iter(self.data)

    def __len__(self):
        """
        Return the number of items in the BaseManager

        :return: number of items in the BaseManager
        :rtype: int
        """
        return len(self.data)

    def __keytransform__(self, key):
        """
        Identity transform for the keys. This function can be reimplemented by
        a subclass to implement a different key transform.

        :param key: key to transform
        :return: the transformed key (in this case the same key as inputted)
        """
        return key

    def _print_key(self, key):
        """
        Return a printable version of the key

        :param key: the key to convert
        :return: printable version of the key
        :rtype: str
        """
        return str(key)

    def _print_value(self, val):
        """
        Return a printable version of the value

        :param key: the value to convert
        :return: printable version of the value
        :rtype: str
        """
        return str(val)

    @abstractproperty
    def _item_extension(self):
        """
        Abstract property that sets the extension of the files to be loaded by
        the BaseManager. When scanning for items, only files with this
        extension will be loaded.

        :return: desired extension
        :rtype: str
        :raises FastrNotImplementedError: if property not reimplemented in subclass
        """
        raise FastrNotImplementedError("Purposefully not implemented!")

    def match_filename(self, filename):
        """
        Check if the filename matches the pattern the manager expects.

        :param filename: filename to match
        :return: flag indicating that the filename matches
        """
        extension = self._item_extension
        if extension.startswith('\\.'):
            pass
        elif extension.startswith('.'):
            extension = '\\{}'.format(extension)
        else:
            extension = '\\.{}'.format(extension)

        pattern = "^.*{}$".format(extension)
        result = re.match(pattern, filename)
        fastr.log.debug('Matching {} with {} -> {}'.format(filename, pattern, result))
        return result is not None

    @abstractmethod
    def _load_item(self, filepath, namespace):
        """
        Abstract method to load an item of the BaseManager. This function is
        not implemented and needs to be reimplemented by a subclass.

        :param str filepath: path of the item to load
        :param str namespace: the namespace of the item to be loaded
        :return: the loaded item
        :raises FastrNotImplementedError: if called without being reimplemented by a subclass
        """
        raise FastrNotImplementedError("Purposefully not implemented!")

    def _store_item(self, name, value):
        """
        Store an item in the BaseManager, will ignore the item if the key is
        already present in the BaseManager.

        :param name: the key of the item to save
        :param value: the value of the item to save
        :return: None
        """
        if name in self.keys():
            fastr.log.warning('Skipping {} from {} (the plugin is already in the {})'.format(name, value.filename, type(self).__name__))
        else:
            fastr.log.debug('Loaded {} {} from {}'.format(type(value).__name__, name, value.filename))
            self[name] = value

    def _scan_directory(self, path, recursive=False, namespace=None):
        """
        Scan a directory for items of the BaseManager to load. All files in a
        directory will be checked. If the extension matches the
        BaseManager._item_extension it will be loaded and added to the
        BaseManager.

        :param str path: path to scan
        :param bool recursive: indicate whether or not to scan the directory recursively
        :return: None
        """
        # Unpack lists, tuples, deques and other objects with a valid __iter__
        if not isinstance(path, (str, unicode)) and isinstance(path, collections.Iterable):
            for entry in path:
                self._scan_directory(entry, recursive)
        else:
            if not os.path.exists(path):
                fastr.log.warning('Cannot scan {} with {}, path does not exist!'.format(path, type(self).__name__))
                return

            if namespace is None:
                namespace = ()

            # Scan directory
            self._scanned_directories.append(path)
            fastr.log.debug('{} scanning {} (recursive={})'.format(type(self).__name__, path, recursive))
            for filename in os.listdir(path):
                filepath = os.path.join(path, filename)

                if os.path.isdir(filepath):
                    if recursive:
                        self._scan_directory(filepath, recursive, namespace + (filename,))
                elif os.path.isfile(filepath) and self.match_filename(filename):
                    # Since we cannot know what Plugins might throw, catch all
                    # pylint: disable=bare-except
                    try:
                        self._load_item(filepath, namespace='.'.join(namespace))
                    except Exception as exception:
                        fastr.log.warning('Could not load file {}: {}'.format(filepath, exception))
                        exc_type, _, _ = sys.exc_info()
                        exc_info = traceback.format_exc()
                        fastr.log.debug('Encountered exception ({}) during loading:\n{}'.format(exc_type.__name__, exc_info))
